#!/usr/bin/env python
# filename: progbar.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import sys
from datetime import datetime


def progress_bar(finished, total, start_time=None):
    '''
    Prints an ASCII progress bar.

    Each call to ``progress_bar`` will update the progress bar. An example
    of tracking the progress of a list of items would look like::

        job_list = [job1, job2, job3, ... jobN]
        total_jobs = len(job_list)

        #initialize the progress bar
        progress_bar(0, total_jobs)

        # do the jobs
        for i, job in enumerate(job_list):
            do_job(job)
            progress_bar(i + 1, total_jobs)

    Args:

        finished (int): Number of finished jobs.

        total (int): Total number of jobs.

        start_time (datetime): Start time, as a ``datetime.datetime`` object.
            Only required if you want to display execution time alongside
            the progress bar. If not provided, execution time is not shown.

    '''
    pct = int(100. * finished / total)
    ticks = pct / 2
    spaces = 50 - ticks
    if start_time is not None:
        elapsed = (datetime.now() - start_time).seconds
        minutes = elapsed / 60
        seconds = elapsed % 60
        minute_str = '0' * (2 - len(str(minutes))) + str(minutes)
        second_str = '0' * (2 - len(str(seconds))) + str(seconds)
        prog_bar = '\r({}/{}) |{}{}|  {}% ({}:{})'.format(finished, total,
            '|' * ticks, ' ' * spaces, pct, minute_str, second_str)
    else:
        prog_bar = '\r({}/{}) |{}{}|  {}%  '.format(finished, total,
            '|' * ticks, ' ' * spaces, pct)
    sys.stdout.write(prog_bar)
    sys.stdout.flush()



#!/usr/bin/env python
# filename: color.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function, division

import seaborn as sns

import numpy as np

from matplotlib import pyplot as plt
from matplotlib import cm, colors


cmaps = {'heatmap': sns.diverging_palette(240, 10, as_cmap=True)}


def cmap_from_color(color, dark=False):
    '''
    Generates a matplotlib colormap from a single color.

    Colormap will be built, by default, from white to ``color``.

    Args:

        color: Can be one of several things:

            1. Hex code
            2. HTML color name
            3. RGB tuple

        dark (bool): If ``True``, colormap will be built from ``color`` to
            black. Default is ``False``, which builds a colormap from
            white to ``color``.

    Returns:

        colormap: A matplotlib colormap

    '''
    if dark:
        return sns.dark_palette(color, as_cmap=True)
    else:
        return sns.light_palette(color, as_cmap=True)


def hex_to_rgb(hex_string):
    rgb = colors.hex2color(hex_string)
    return tuple([int(255 * x) for x in rgb])


def rgb_to_hex(rgb_tuple):
    div = 1 if all([v <= 1.0 for v in rgb_tuple]) else 255
    return colors.rgb2hex([1.0 * x / div for x in rgb_tuple])


def hls(n_colors, hue=0.01, lightness=0.6, saturation=0.65):
    return sns.hls_palette(n_colors, h=hue, l=lightness, s=saturation)


def husl(n_colors, hue=0.01, saturation=0.9, lightness=0.65):
    return sns.husl_palette(n_colors, h=hue, s=saturation, l=lightness)


def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=256):
    """
    Truncates a colormap, such that the new colormap consists of
    ``cmap[minval:maxval]``.

    If maxval is larger than minval, the truncated colormap will be reversed.

    Args:

    	cmap (colormap): Colormap to be truncated

    	minval (float): Lower bound. Should be a float betwee 0 and 1.

    	maxval (float): Upper bound. Should be a float between 0 and 1

    	n (int): Number of colormap steps. Default is ``256``.

    Returns:

    	colormap: A matplotlib colormap

    http://stackoverflow.com/questions/18926031/how-to-extract-a-subset-of-a-colormap-as-a-new-colormap-in-matplotlib
    """
    cmap = get_cmap(cmap)
    name = "%s-trunc-%.2g-%.2g" % (cmap.name, minval, maxval)
    return colors.LinearSegmentedColormap.from_list(
        name, cmap(np.linspace(minval, maxval, n)))


def stack_colormap(lower, upper, n=256):
    """
    Stacks two colormaps (``lower`` and ``upper``) such that
    low half -> ``lower`` colors, high half -> ``upper`` colors

    Args:

    	lower (colormap): colormap for the lower half of the stacked colormap.

    	upper (colormap): colormap for the upper half of the stacked colormap.

    	n (int): Number of colormap steps. Default is ``256``.
    """
    A = get_cmap(lower)
    B = get_cmap(upper)
    name = "%s-%s" % (A.name, B.name)
    lin = np.linspace(0, 1, n)
    return array_cmap(np.vstack((A(lin), B(lin))), name, n=n)


def get_cmap(cmap=None, name=None, from_color=None, dark=False, n=256):
    # """
    # Generates a matplotlib colormap.

    # cmap can be one of several things:
    #     - a name ('Blues', 'BuGn_r') of a built-in colormap
    #     - a cmap
    #     - a filename, np.loadtxt() n x 3 or 4  ints 0..255 or floats 0..1
    #     - a numpy array
    # See http://wiki.scipy.org/Cookbook/Matplotlib/Show_colormaps or in IPython, plt.cm.<tab>

    # An optional name for the colormap can be provided (name), as well as the number
    # of cmap steps (n).

    # Alternatively, to make a colormap using a single color, provide the color
    # via from_color. By default, the supplied color will be the dark end of
    # the cmap, with white as the lightest color. To reverse (use the input
    # color as the lighest value and black as the darkest), set dark=True.

    # Returns a matplotlib colormap object.
    # """
    if from_color is not None:
        return cmap_from_color(from_color, dark)
    elif cmap is None:
        err = 'You must provide either cmap or from_color'
        raise RuntimeError(err)
    if isinstance(cmap, colors.Colormap):
        return cmap
    if isinstance(cmap, basestring):
        if cmap in cm.cmap_d:
            return plt.get_cmap(cmap)  # "Blues" ...
        A = np.loadtxt(cmap, delimiter=None)  # None: white space
        name = name or cmap.split("/")[-1].split(".")[0]  # .../xx.csv -> xx
    else:
        A = cmap  # numpy array or array-like
    return array_cmap(A, name, n=n)



#!/usr/bin/python
# filename: timepoint.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


class Timepoint(object):
    """
    Stores and manipulates timepoint information.
    """
    def __init__(self, tp, order, color):
        super(Timepoint, self).__init__()
        self.name = tp
        self.order = int(order)
        self.raw_color = color
        self.color = self._get_color()


    def _get_color(self):
        if self._is_rbg():
            return self._convert_to_hex()
        return self.raw_color


    def _is_rbg(self):
        color = self.raw_color
        if type(color) == tuple:
            return True
        elif len(self.raw_color.split(',')) == 3:
            self.raw_color = self._convert_to_tuple()
            return True
        return False


    def _convert_to_tuple(self):
        c = self.raw_color.split(',')
        r = float(c[0].replace('(', '').strip())
        g = float(c[1].strip())
        b = float(c[2].replace(')', '').strip())
        return (r, g, b)


    def _convert_to_hex(self):
        r = float(self.raw_color[0])
        g = float(self.raw_color[1])
        b = float(self.raw_color[2])
        if 0 < round(sum([r, b, g]), 2) <= 3:
            r, b, g = (255 * r, 255 * b, 255 * g)
        return '#%02x%02x%02x' % (r, b, g)



#!/usr/bin/python
# filename: msa.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import re
import subprocess as sp


def align(seqs, fasta):
    '''
    Makes a multiple sequence alignment.

    Input
    seqs: sequences to be aligned (as biopython SeqIO objects)
    fasta: path to use for the alignment input file

    Output
    path to the alignment file generated by MUSCLE
    '''
    build_alignment_input(seqs, fasta)
    return muscle(fasta, len(seqs))


def build_alignment_input(seqs, fasta):
    flist = []
    for s in seqs:
        seq_id = re.sub('[:;]', '_', s.id)
        flist.append('>{}\n{}'.format(seq_id, str(s.sequence).upper()))
    open(fasta, 'w').write('\n'.join(flist))


def muscle(fasta, count):
    output_dir = os.path.dirname(fasta)
    output_name = '{}_aligned.aln'.format(os.path.basename(fasta).replace('.fasta', ''))
    output = os.path.join(output_dir, output_name)
    if count < 100:
        muscle_cline = 'muscle -in {} -out {}'.format(fasta, output)
    elif count < 1000:
        muscle_cline = 'muscle -maxiters 2 -in {} -out {}'.format(fasta, output)
    else:
        muscle_cline = 'muscle -maxiters 1 -diags -in {} -out {}'.format(fasta, output)
    m = sp.Popen(muscle_cline, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = m.communicate()
    return output



#!/usr/bin/env python
# filename: cluster.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import logging
import multiprocessing as mp
import os
import random
import sqlite3
import string
import subprocess as sp
import sys
import tempfile
import time

from Bio import SeqIO, AlignIO
from Bio.Align import AlignInfo

from abtools import log
from abtools.alignment import mafft
from abtools.sequence import Sequence
from abtools.utils.decorators import lazy_property


class Cluster(object):
    """
    Data and methods for a cluster of sequences.

    All public attributes are evaluated lazily, so attributes that
    require significant processing time are only computed when needed.
    In addition, attributes are only calculated once, so if you
    change the Cluster object after accessing attributes, the
    attributes will not update. Setters are provided for all attributes,
    however, so you can update them manually if necessary::

        seqs = [Sequence1, Sequence2, ... SequenceN]
        clust = cluster(seqs)

        # calculate the consensus
        consensus = clust.consensus

        # add sequences to the Cluster
        more_sequences = [SequenceA, SequenceB, SequenceC]
        clust.sequences += more_sequences

        # need to recompute the consensus manually
        clust.consensus = clust._make_consensus()


    Attributes:

        ids (list): A list of all sequence IDs in the Cluster

        size (int): Number of sequences in the Cluster

        sequences (list): A list of all sequences in the Cluster,
            as AbTools ``Sequence`` objects.

        consensus (Sequence): Consensus sequence, calculated by
            aligning all sequences with MAFFT and computing the
            ``Bio.Align.AlignInfo.SummaryInfo.gap_consensus()``

        centroid (Sequence): Centroid sequence, as calculated by
            CD-HIT.
    """
    def __init__(self, raw_cluster, seq_db=None, db_path=None, seq_dict=None):
        super(Cluster, self).__init__()
        self._raw_cluster = raw_cluster
        self._seq_db = seq_db
        self._seq_db_path = db_path
        self._seq_dict = seq_dict


    @lazy_property
    def ids(self):
        return self._get_ids()

    @lazy_property
    def size(self):
        return len(self.ids)

    @lazy_property
    def sequences(self):
        if all([self._seq_db is None, self._seq_dict is None]):
            err = "ERROR: In order to access a Cluster's sequences, you must provide "
            err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
            raise RuntimeError(err)
        return self._get_sequences()

    @lazy_property
    def consensus(self):
        if all([self._seq_db is None, self._seq_dict is None]):
            err = "ERROR: In order to compute a Cluster's consensus, you must provide "
            err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
            raise RuntimeError(err)
        return self._make_consensus()

    @lazy_property
    def centroid(self):
        if all([self._seq_db is None, self._seq_dict is None]):
            err = "ERROR: In order to retrieve a Cluster's centroid, you must provide "
            err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
            raise RuntimeError(err)
        return self._get_centroid()


    def cleanup(self):
        try:
            os.unlink(self._raw_cluster)
        except:
            pass
        self.terminate_db()


    def terminate_db(self):
        if self._seq_db is not None:
            self._seq_db.close()
        if self._seq_db_path is not None:
            if os.path.isfile(self._seq_db_path):
                os.unlink(self._seq_db_path)


    def _get_ids(self):
        ids = []
        for c in self._raw_cluster[1:]:
            if c:
                ids.append(c.split()[2][1:-3])
        return ids

    def _get_sequences(self):
        if self._seq_db is not None:
            seqs = []
            for chunk in self._chunker(self.ids):
                sql_cmd = '''SELECT seqs.id, seqs.sequence
                             FROM seqs
                             WHERE seqs.id IN ({})'''.format(','.join('?' * len(chunk)))
                seq_chunk = self._seq_db.execute(sql_cmd, chunk)
                seqs.extend(seq_chunk)
            return [Sequence(s) for s in seqs]
        else:
            return [self._seq_dict[s] for s in self.ids]

    def _get_centroid(self):
        for line in self._raw_cluster:
            if '*' in line:
                centroid_id = line.split()[2][1:-3]
                break
        centroid = [s for s in self.sequences if s.id == centroid_id][0]
        return centroid

    def _make_consensus(self):
        if len(self.sequences) == 1:
            return self.sequences[0]
        _aln = mafft(self.sequences, as_file=True)
        aln = AlignIO.read(open(_aln, 'r'), 'fasta')
        summary_align = AlignInfo.SummaryInfo(aln)
        consensus = summary_align.gap_consensus(threshold=0.51, ambiguous='n')
        consensus_string = str(consensus).replace('-', '')
        consensus_seq = Sequence(consensus_string.upper())
        os.unlink(_aln)
        return consensus_seq

    @staticmethod
    def _chunker(l, size=900):
        return (l[pos:pos + size] for pos in xrange(0, len(l), size))


def cluster(seqs, threshold=0.975, out_file=None, make_db=True, temp_dir=None,
            quiet=False, threads=0, return_just_seq_ids=False, max_memory=800, debug=False):
    '''
    Perform sequence clustering with CD-HIT.

    Args:

        seqs (list): An iterable of sequences, in any format that abtools.sequence.Sequence()
            can handle

        threshold (float): Clustering identity threshold. Default is 0.975.

        out_file (str): Path to the clustering output file. Default is to use
            tempfile.NamedTempraryFile to generate an output file name.

        temp_dir (str): Path to the temporary directory. If not provided, '/tmp' is used.

        make_db (bool): Whether to build a SQlite database of sequence information. Required
            if you want to calculate consensus/centroid sequences for the resulting
            clusters or if you need to access the clustered sequences (not just sequence IDs)
            Default is True.

    Returns:

        list: A list of Cluster objects, one per cluster.
    '''
    if make_db:
        ofile, cfile, seq_db, db_path = cdhit(seqs, out_file=out_file, temp_dir=temp_dir,
            threshold=threshold, make_db=True, quiet=quiet, threads=threads, max_memory=max_memory, debug=debug)
        return parse_clusters(ofile, cfile, seq_db=seq_db, db_path=db_path, return_just_seq_ids=return_just_seq_ids)
    else:
        seqs = [Sequence(s) for s in seqs]
        seq_dict = {s.id: s for s in seqs}
        ofile, cfile, = cdhit(seqs, out_file=out_file, temp_dir=temp_dir, threads=threads,
            threshold=threshold, make_db=False, quiet=quiet, max_memory=max_memory, debug=debug)
        return parse_clusters(ofile, cfile, seq_dict=seq_dict, return_just_seq_ids=return_just_seq_ids)


def cdhit(seqs, out_file=None, temp_dir=None, threshold=0.975, make_db=True, quiet=False, threads=0, max_memory=800, debug=False):
    # '''
    # Perform CD-HIT clustering on a set of sequences.

    # Inputs are an iterable of sequences, which can be in any format that abtools.sequence.Sequence
    # can handle.

    # Returns the centroid file name and cluster file name (from CD-HIT).
    # If ::make_db:: is True (default), a SQLite3 connection and database path are also returned.
    # '''
    logger = log.get_logger('cluster')
    start_time = time.time()
    seqs = [Sequence(s) for s in seqs]
    if not quiet:
        logger.info('CD-HIT: clustering {} seqeunces'.format(len(seqs)))
    if out_file is None:
        out_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
        ofile = out_file.name
    else:
        ofile = os.path.expanduser(out_file)
    ifile = _make_cdhit_input(seqs, temp_dir)
    cdhit_cmd = 'cd-hit -i {} -o {} -c {} -n 5 -d 0 -T {} -M {}'.format(ifile,
                                                                        ofile,
                                                                        threshold,
                                                                        threads,
                                                                        max_memory)
    cluster = sp.Popen(cdhit_cmd,
                       shell=True,
                       stdout=sp.PIPE,
                       stderr=sp.PIPE)
    stdout, stderr = cluster.communicate()
    if debug:
        print(stdout)
        print(stderr)
    else:
        os.unlink(ifile)
    if not quiet:
        logger.info('CD-HIT: clustering took {:.2f} seconds'.format(time.time() - start_time))
    cfile = ofile + '.clstr'
    if make_db:
        if not quiet:
            logger.info('CD-HIT: building a SQLite3 database')
        seq_db, db_path = _build_seq_db(seqs, direc=temp_dir)
        return ofile, cfile, seq_db, db_path
    return ofile, cfile


def parse_clusters(out_file, clust_file, seq_db=None, db_path=None, seq_dict=None, return_just_seq_ids=False):
    # '''
    # Parses clustered sequences.

    # Inputs are a CD-HIT cluster file (ends with '.clstr') and a connection to a
    # SQLite3 database of sequence IDs and sequences.

    # Returns a list of Cluster objects (one per cluster).
    # '''
    raw_clusters = [c.split('\n') for c in open(clust_file, 'r').read().split('\n>')]
    if return_just_seq_ids:
        ids = []
        for rc in raw_clusters:
            _ids = []
            for c in rc[1:]:
                if c:
                    _ids.append(c.split()[2][1:-3])
            ids.append(_ids)
        os.unlink(out_file)
        os.unlink(clust_file)
        return ids
    os.unlink(out_file)
    os.unlink(clust_file)
    return [Cluster(rc, seq_db, db_path, seq_dict) for rc in raw_clusters]


def _make_cdhit_input(seqs, temp_dir):
    ifile = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
    fastas = [s.fasta for s in seqs]
    ifile.write('\n'.join(fastas))
    ifile.close()
    return ifile.name






def _build_seq_db(seqs, direc=None):
    # '''
    # Builds a SQLite3 database of sequences.

    # Inputs are a list of Sequence objects and an optional directory to store the database.
    # If ::direc:: is not provided, '/tmp' will be used.

    # Returns a SQLite3 connection object and the database path.
    # '''
    direc = direc if direc is not None else '/tmp'
    db_name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
    db_path = os.path.join(direc, db_name)
    conn = sqlite3.connect(db_path)
    c = conn.cursor()
    create_cmd = '''CREATE TABLE seqs (id text, sequence text)'''
    insert_cmd = 'INSERT INTO seqs VALUES (?,?)'
    c.execute('DROP TABLE IF EXISTS seqs')
    c.execute(create_cmd)
    c.executemany(insert_cmd, [(str(s.id), str(s.sequence)) for s in seqs])
    c.execute('CREATE INDEX seq_index ON seqs (id)')
    return c, db_path



#!/usr/local/bin/python
# filename: databases.py

#
# Copyright (c) 2017 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import cPickle as pickle
import os
import sqlite3
import time


class KeyValueDB(object):
    """A SQLite database for storing key/value pairs.

    Args:

        name (str): Name for the database file. Required, except for in-memory databases.

        dir (str): Directory for the database. Required, except for in-memory databases.

        pickle_values (bool): If ``True``, values will be pickled (with ``cPickle``) before
            entry into the database. Necessary if you want to store Python objects as values.
            Default is ``False``.

        in_memory (bool): If ``True``, the KeyValueDB will be stored in memory rather than on disk.
            Default is ``False``.

    """
    def __init__(self, name=None, dir=None, pickle_values=True, in_memory=False):
        super(Database, self).__init__()
        self._name = name
        self._dir = dir
        self._path = None
        self.table_name = 'data'
        self.pickle_values = pickle_values
        self.structure = [('key', 'text'), ('value', 'text')]
        self.initialized = False
        self._connection = None
        self._cursor = None
        self._create_table_cmd = None
        self._insert_cmd = None
        if not self.initialized:
            self.create_table()

    def __getitem__(self, key):
        return self.find_one(key)

    def __setitem__(self, key, value):
        return self.insert_one(key, value)


    @property
    def name(self):
        if self._name is None:
            self._name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
        return self._name

    @name.setter
    def name(self, name):
        self._name = name


    @property
    def dir(self):
        if self._dir is None:
            if self.in_memory:
                self._dir = None
            else:
                self._dir = '/tmp'
        return self._dir

    @dir.setter
    def dir(self, dir):
        self._dir = dir


    @property
    def path(self):
        if self._path is None:
            if self.in_memory:
                self._path = ':memory:'
            else:
                self._path = os.path.join(self.dir, self.name)
        return self._path


    @property
    def connection(self):
        if self._connection is None:
            self._connection = sqlite3.connect(self.path)
        return self._connection

    @connection.setter
    def connection(self, connection):
        self._connection = connection


    @property
    def cursor(self):
        if self._cursor is None:
            self._cursor = self.connection.cursor()
        return self._cursor

    @cursor.setter
    def cursor(self, cursor):
        self._cursor = cursor


    @property
    def create_table_cmd(self):
        if self._create_table_cmd is None:
            field_string = ', '.join([' '.join(s) for s in self.structure])
            self._create_table_cmd = 'CREATE TABLE {} ({})'.format(self.table_name,
                field_string)
        return self._create_table_cmd


    @property
    def insert_cmd(self):
        if self._insert_cmd is None:
            self._insert_cmd = 'INSERT INTO {} VALUES ({})'.format(self.table_name,
                ','.join(['?'] * len(self.structure)))
        return self._insert_cmd


    def commit(self):
        '''
        Commits changes to the database.
        '''
        self.connection.commit()


    def close(self):
        '''
        Closes the database connection, but leaves the KeyValueDB object intact.
        The connection will automatically be re-established if necessary. Primarily designed
        to allow the KeyValueDB object to be passed as an argument via ``multiprocessing``,
        as open SQLite connection objects cannot be pickled.
        '''
        self.connection.close()
        del self._cursor
        del self._connection
        self._cursor = None
        self._connection = None


    def terminate(self):
        '''
        Terminates the database and, for databases stored on disk, removes the database file.
        All data will be lost and reconnection will not be possible.
        '''
        self.close()
        if not self.in_memory:
            os.unlink(self.path)


    def create_table(self):
        self.cursor.execute('DROP TABLE IF EXISTS {}'.format(self.table_name))
        self.cursor.execute(self.create_table_cmd)
        self.initialized = True


    def insert_one(self, key, value):
        '''
        Inserts a single key/value pair

        Args:

            key (str): The key

            value: The value. Can be an arbitrary object if the database was initialised
                with ``pickle_values``, otherwise should be a ``str``, ``int`` or ``float``.
        '''
        if self.pickle_values:
            value = pickle.dumps(value, protocol=0)
        with self.connection as conn:
            conn.execute(self.insert_cmd, (key, value))


    def insert_many(self, data):
        '''
        Inserts multiple key/value pairs.

        Args:

            data: a list/generator of iterables (lists or tuples) with each iterable containing
                a single key/value pair.
        '''
        if self.pickle_values:
            data = ((d[0], pickle.dumps(d[1], protocol=0)) for d in data)
        with self.connection as conn:
            conn.executemany(self.insert_cmd, data)


    def find_one(self, key):
        '''
        Finds one value with a key that matches ``key``.

        Args:

            key (str): a single key

        Returns:

            a single value (unpickled if necessary)
        '''
        self.cursor.execute(
            '''SELECT data.key, data.value
               FROM data
               WHERE data.key LIKE ?''', (key, ))
        if self.pickle_values:
            return pickle.loads(str(self.cursor.fetchone()[1]))
        return self.cursor.fetchone()[1]


    def find(self, keys):
        '''
        Finds all values with a key that match ``keys``. If ``keys`` is a list, values from
        any records with a key that matches any value in ``keys`` will be returned.

        Args:

            keys: a single key (as a ``str``) or iterable (``list`` or ``tuple``)
                containing one or more keys

        Returns:

            list: values, unpickled if necessary
        '''
        if type(keys) in [str, unicode]:
            keys = [keys, ]
        results = []
        for chunk in self.chunker(keys):
            result_chunk = self.cursor.execute(
                '''SELECT data.key, data.value
                   FROM data
                   WHERE data.key IN ({})'''.format(','.join('?' * len(chunk))), chunk)
            results.extend(result_chunk)
        if self.pickle_values:
            return [pickle.loads(str(r[1])) for r in results]
        return [r[1] for r in results]


    def find_all(self):
        '''
        Returns all values in the database.

        Returns:

            list: all values, unpickled if necessary
        '''
        results = self.cursor.execute(
            '''SELECT data.value
               FROM data''')
        if self.pickle_values:
            return [pickle.loads(str(r[0])) for r in results]
        return [r[0] for r in results]


    def delete(self, keys):
        '''
        Deletes all records with a key that match ``keys``. If ``keys`` is a list, any records
        with a key that matches any value in ``keys`` will be deleted.

        Args:

            keys: a single key (as a ``str``) or iterable (``list`` or ``tuple``)
                containing one or more keys
        '''
        if type(keys) in [str, unicode]:
            keys = [keys, ]
        with self.connection as conn:
            conn.executemany(
                '''DELETE
                   FROM data
                   WHERE data.key == ?''', keys)


    def index(self, field='key'):
        '''
        Indexes the database

        Args:

            field (str): The field on which to create the index. Default is 'key'.
        '''
        index_name = field + '_index'
        self.cursor.execute('CREATE INDEX {} ON {} ({})'.format(index_name,
            self.table_name,
            field))


    @staticmethod
    def chunker(l, n=900):
        '''
        Yields successive n-sized chunks from l.
        '''
        for i in xrange(0, len(l), n):
            yield l[i:i + n]



#!/usr/bin/python
# filename: _compare.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

from collections import Counter
import itertools
import math
import multiprocessing as mp
import os
import random
import sqlite3
from StringIO import StringIO
import subprocess as sp
import sys
import tempfile
import time
import urllib
import uuid

import numpy as np
import pandas as pd

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

from abtools import log, mongodb



def parse_args():
    import argparse
    parser = argparse.ArgumentParser("Compares two or more antibody repertoires.")
    parser.add_argument('-o', '--output', dest='output', required=True,
                        help="Output directory for figure and data files. Required.")
    parser.add_argument('-l', '--log', dest='log',
                        help="Location for the log file. \
                        If not provided, log will be written to <output>/abcompare.log")
    parser.add_argument('-d', '--database', dest='db', required=True,
                        help="Name of the MongoDB database to query. Required.")
    parser.add_argument('-1', '--collection1', dest='collection1', default=None,
                        help="Name of the first MongoDB collection to query for comparison. \
                        If -1 and -2 are not provided, all collections in the given database will be processed iteratively.")
    parser.add_argument('-2', '--collection2', dest='collection2', default=None,
                        help="Name of the second MongoDB collection to query for comparison. \
                        If not provided, the collection provided by -1 will be iteratively compared with all other collections in the database. \
                        If both -1 and -2 are not provided, all collections in the given database will be processed iteratively.")
    parser.add_argument('--collection-prefix', dest='collection_prefix', default=None,
                        help="If supplied, will iteratively process only collections beginning with <collection_prefix>.\
                        if '-1' is also provided, it will be iteratively compared with all collections beginning with <collection_prefix>.")
    parser.add_argument('-i', '--ip', dest='ip', default='localhost',
                        help="IP address for the MongoDB server.  Defaults to 'localhost'.")
    parser.add_argument('--port', dest='port', default=27017, type=int,
                        help="Port for the MongoDB server.  Defaults to 27017.")
    parser.add_argument('-u', '--user', dest='user', default=None,
                        help="Username for the MongoDB server. Not used if not provided.")
    parser.add_argument('-p', '--password', dest='password', default=None,
                        help="Password for the MongoDB server. Not used if not provided.")
    parser.add_argument('-k', '--chunksize', dest='chunksize', type=int, default=100000,
                        help="Number of sequences to use for each similarity calculation. \
                        Default is 100,000.")
    parser.add_argument('-I', '--iterations', dest='iterations', type=int, default=1000,
                        help="Number of iterations of the similarity calculation to perform. \
                        Default is 10000.")
    parser.add_argument('-s', '--similarity-method', dest='method',
                        choices=['marisita-horn', 'kullback-leibler', 'jensen-shannon', 'jaccard', 'bray-curtis', 'renkonen', 'cosine'],
                        default='marisita-horn',
                        help="The similarity calculation to use. \
                        Options are 'marisita-horn', 'kullback-leibler' and 'jensen-shannon'. \
                        Note that kullback-leibler is actually a divergence measure (lower values indicate greater similarity) \
                        Default is 'marisita-horn'.")
    parser.add_argument('-c', '--control', dest='control_similarity', default=False, action='store_true',
                        help="Plot the control similarity of the two collections")
    parser.add_argument('-C', '--chain', dest='chain', default='heavy', choices=['heavy', 'kappa', 'lambda'],
                        help="Select the antibody chain to analyze. \
                        Options are 'heavy', 'kappa', and 'lambda'. \
                        Default is 'heavy'.")
    parser.add_argument('--debug', dest='debug', action='store_true', default=False,
                        help="If set, will run in debug mode.")
    return parser


class Args(object):
    def __init__(self, output=None, log=None, db=None,
                 collection1=None, collection2=None, collection_prefix=None,
                 ip='localhost', port=27017, user=None, password=None,
                 chunksize=100000, iterations=10000,
                 method='marisita-horn', control_similarity=False,
                 chain='heavy', debug=False):
        super(Args, self).__init__()
        if not all([output, db]):
            logging.info("\nERROR: Ouput directory and MongoDB database name must be provided.\n")
            sys.exit(1)
        self.output = output
        args.log = log
        self.db = db
        self.collection1 = collection1
        self.collection2 = collection2
        self.collection_prefix = collection_prefix
        self.ip = str(ip)
        self.port = int(port)
        self.user = user
        self.password = password
        self.chunksize = int(chunksize)
        self.iterations = int(iterations)
        self.method = method
        self.control_similarity = control_similarity
        if chain not in ['heavy', 'kappa', 'lambda']:
            logging.info('\nERROR: Please select an appropriate chain. \
                Valid choices are: heavy, kappa and lambda.\n')
            sys.exit(1)
        self.chain = chain
        self.debug = debug



# ================================================
#
#                   MONGO
#
# ================================================



def get_collection_pairs(db, args):
    if args.collection1 and args.collection2:
        print_single_pair_info(args.collection1, args.collection2)
        return [(args.collection1, args.collection2), ]
    collections = db.collection_names(include_system_collections=False)
    if args.collection_prefix:
        collections = [c for c in collections if c.startswith(args.collection_prefix)]
    if args.collection1:
        collections.sort()
        pairs = zip([args.collection1] * len(collections), collections)
        print_multiple_pairs_info(pairs)
        return pairs
    pairs = [p for p in itertools.combinations(sorted(collections), 2)]
    print_multiple_pairs_info(pairs)
    return pairs


def index_collections(db, pairs):
    collections = sorted(list(set([c for tup in pairs for c in tup])))
    fields = ['chain', 'prod']
    logger.info('')
    logger.info('Creating indexes...')
    for c in collections:
        print(c)
        mongodb.index(db, c, fields)


def query(db, collection, chain):
    logger.info('Getting {} sequences from MongoDB...'.format(collection))
    sys.stdout.flush()
    c = db[collection]
    results = c.find({'chain': chain, 'prod': 'yes'},
                     {'_id': 0, 'v_gene.gene': 1, 'cdr3_len': 1})
    seqs = [r for r in results if '/OR' not in r['v_gene']['gene']]
    logger.info('Found {} sequences'.format(len(seqs)))
    return seqs



# ================================================
#
#                   MUNGING
#
# ================================================



def random_sample_no_replacement(s1, s2, only_continuous, norm, args):
    chunksize = min(int(0.9 * len(s1)), int(0.9 * len(s2)), args.chunksize)
    c1_indexes = []
    c2_indexes = []
    c1 = random.sample(s1, chunksize)
    c2 = random.sample(s2, chunksize)
    agg1 = aggregate(c1)
    agg2 = aggregate(c2)
    df = pd.DataFrame({'s1': agg1, 's2': agg2})
    df.fillna(0, inplace=True)
    if only_continuous:
        df = df[(df['s1'] != 0) & (df['s2'] != 0)]
    if norm:
        return normalize(df['s1'], df['s2'])
    return df['s1'], df['s2']


def get_vgenes(db, c1, chain, prev_data=None):
    data1 = query(db, c1, chain)
    v1 = [d['v_gene']['gene'] for d in data1]
    if prev_data:
        size = min(len(prev_data), len(v1))
        v1 = v1[:size]
        prev_data = prev_data[:size]
        return prev_data, v1
    return v1


def aggregate(data):
    '''
    Counts the number of occurances of each item in 'data'.

    Input
    data: a list of values.

    Output
    a dict of bins and counts.
    '''
    vdict = {}
    for d in data:
        vdict[d] = vdict[d] + 1 if d in vdict else 1
    return vdict


def normalize(s1, s2):
    return s1 / np.sum(s1), s2 / np.sum(s2)





# ================================================
#
#                 SIMILARITIES
#
# ================================================



def simdif_method(sample1, sample2, args):
    # '''
    # Determines the appropriate similarity/divergence method to use
    # based on user options.

    # Inputs
    # sample1: list of properties (i.e. Var genes) for sample 1
    # sample2: list of properties (i.e. Var genes) for sample 2

    # Outputs (from appropriate method function)
    # median: median of similarity/divergence score distribution
    # counts: from binned scores, counts for each bin
    # bins: from binned scores, left-most boundary of each bin
    # similarities: raw similarity scores.
    # '''
    methods = {'marisita-horn': (mh_similarity, False, False),
               'kullback-leibler': (kl_divergence, True, True),
               'jensen-shannon': (js_similarity, False, True),
               'jaccard': (jaccard_similarity, False, False),
               'bray-curtis': (bc_similarity, False, True),
               'renkonen': (renkonen_similarity, False, True),
               'cosine': (cosine_similarity, False, False)}
    method, continuous, normalize = methods[args.method]
    s1, s2 = random_sample_no_replacement(sample1,
                                          sample2,
                                          continuous,
                                          normalize,
                                          args)
    return method(s1, s2)


def calculate_similarities(s1, s2, args):
    if args.debug:
        similarities = []
        for i in xrange(args.iterations):
            similarities.append(simdif_method(s1, s2, args))
    else:
        async_results = []
        p = mp.Pool()
        for i in xrange(args.iterations):
            async_results.append(p.apply_async(simdif_method, (s1, s2, args)))
        monitor_mp_jobs(async_results, sys.stdout)
        similarities = [a.get() for a in async_results]
    median = np.median(similarities)
    counts, bins = np.histogram(similarities)
    return median, counts, bins, similarities


def calculate_control_similarities(s1, s2, args):
    pool = s1 + s2
    random.shuffle(pool)
    return calculate_similarities(pool, pool, args)


def mh_similarity(sample1, sample2):
    '''
    Calculates the Marista-Horn similarity for two samples.

    .. note:

        sample1 and sample2 should be the same length, and
        the sum of each sample should be greater than 0.

    Args:

        sample1: list of frequencies for sample 1
        sample2: list of frequencies for sample 2

    Returns:

        float: Marista-Horn similarity (between 0 and 1)
    '''
    X = sum(sample1)
    Y = sum(sample2)
    XY = X * Y
    sumXiYi = 0
    sumXiSq = 0
    sumYiSq = 0
    for x, y in zip(sample1, sample2):
        sumXiYi += x * y
        sumXiSq += x * x
        sumYiSq += y * y
    num = 2 * sumXiYi
    denom = (float(sumXiSq) / (X * X) + float(sumYiSq) / (Y * Y)) * XY
    return 1. * num / denom


def kl_divergence(s1, s2):
    '''
    Calculates the Kullback-Leibler divergence for two samples.

    .. note:

        s1 and s2 should be the same length, and the sum of
        each should equal 1. Probabilities should be continuous for
        both s1 and s2.

    Args:

        sample1: probability distribution for sample 1
        sample2: probability distribution for sample 2

    Returns:

        float: Kullbeck-Leibler similarity
    '''
    n1, n2 = normalize(s1, s2)
    kl = 0
    for Pi, Qi in zip(n1, n2):
        kl += Pi * math.log(Pi / Qi)
    return kl


def js_similarity(s1, s2):
    '''
    Calculates the Jensen-Shannon similarity for two samples.

    .. note:

        s1 and s2 should be the same length, and the sum of
        each should equal 1.

    Args:

        sample1: probability distribution for sample 1
        sample2: probability distribution for sample 2

    Returns:

        float: Jensen-Shannon similarity (between 0 and 1)
    '''
    n1, n2 = normalize(s1, s2)
    Hsum = np.zeros(len(n1))
    sumH = 0
    for s in [n1, n2]:
        sumH += 0.5 * shannon_entropy(s)
        for i, val in enumerate(s):
            Hsum[i] += 0.5 * val
    return 1 - (shannon_entropy(Hsum) - sumH)


def shannon_entropy(prob_dist):
    '''
    Calculates the Shannon entropy for a single probability distribution.

    Args:

        prob_dist: probability distribution, must sum to 1

    Returns:

        float: Shannon entropy
    '''
    return -sum([p * math.log(p) for p in prob_dist if p != 0])


def jaccard_similarity(s1, s2):
    '''
    Calculates the Jaccard similarity for two samples.

    .. note:

        sample1 and sample2 should be the same length, and
        the sum of each sample should be greater than 0.

    Args:

        sample1: list of frequencies for sample 1
        sample2: list of frequencies for sample 2

    Returns:

        float: Jaccard similarity (between 0 and 1)
    '''
    num = 0
    denom = 0
    for xy in zip(s1, s2):
        num += min(xy)
        denom += max(xy)
    return 1. * num / denom


def renkonen_similarity(s1, s2):
    '''
    Calculates the Renkonen similarity (also known as the
    percentage similarity) for two samples.

    .. note:

        s1 and s2 should be the same length, and
        the sum of each sample should equal 1.

    Args:

        s1: probability distribution for sample 1
        s2: probability distribution for sample 2

    Returns:

        float: Renkonen similarity (between 0 and 1)
    '''
    return sum(min(vals) for vals in zip(s1, s2))


def bc_similarity(s1, s2):
    '''
    Calculates the Bray-Curtis similarity for two samples.

    .. note:

        s1 and s2 should be the same length, and
        the sum of each sample should equal 1.

    Args:

        s1: probability distribution for sample 1
        s2: probability distribution for sample 2

    Returns:

        float: Bray-Curtis similarity (between 0 and 1)
    '''
    return 2.0 * sum(min(vals) for vals in zip(s1, s2)) / (sum(s1) + sum(s2))


def cosine_similarity(s1, s2):
    '''
    Calculates the cosine (angular) similarity for two samples.

    .. note:

        s1 and s2 should be the same length.

    Args:

        s1: list of frequencies for sample 1
        s2: list of frequencies for sample 2

    Returns:

        float: Cosine similarity (between 0 and 1)
    '''
    num = sum([Pi * Qi for Pi, Qi in zip(s1, s2)])
    denomPi = mp.sqrt(sum([Pi * Pi for Pi in s1])) * mp.sqrt(sum([Qi * Qi for Qi in s2]))
    return float(num) / denom


def sd_similarity(s1, s2):
    '''
    Calculates the Brey-Curtis similarity for two samples.

    .. note:

        s1 and s2 should be the same length, and
        the sum of each sample should equal 1.

    Args:

        s1: list of frequencies for sample 1
        s2: list of frequencies for sample 2

    Results:

        float: Brey-Curtis similarity (between 0 and 1)
    '''
    num = 0
    denom = 0
    for x, y in zip(s1, s2):
        num += min(x, y) > 0
        denom += min(x) > 0
        denom += min(y) > 0
    return 1. * num / denom


def bin_similarities(similarities):
    return np.histogram(similarities, bins=10)


def monitor_mp_jobs(results, log):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        ready = [ar for ar in results if ar.ready()]
        finished = len(ready)
        update_progress(finished, jobs, log)
    log.write('\n\n')


def update_progress(finished, jobs, log, failed=None):
    pct = int(100. * finished / jobs)
    ticks = pct / 2
    spaces = 50 - ticks
    if failed:
        prog_bar = '\r({}/{}) |{}{}|  {}% ({}, {})'.format(finished, jobs, '|' * ticks, ' ' * spaces, pct, finished - failed, failed)
    else:
        prog_bar = '\r({}/{}) |{}{}|  {}%'.format(finished, jobs, '|' * ticks, ' ' * spaces, pct)
    sys.stdout.write(prog_bar)
    sys.stdout.flush()





# ================================================
#
#                OUTPUT FILES
#
# ================================================



def make_sim_plot(counts, bins, median, ofile, output_dir):
    sns.set_style('white')
    bin_mdpt = (bins[1] - bins[0]) / 2
    x = [str(round(b + bin_mdpt, 4)) for b in bins[:-1]]
    y = counts
    # set bar locations and width
    ind = np.arange(len(x))
    width = 0.75
    # plot objects
    fig = plt.figure()
    ax = fig.add_subplot(111)
    # plot aesthetics
    color = sns.hls_palette(7)[4]
    # axis limits, labels and ticks
    ax.set_ylim(0, 1.1 * max(y))
    ax.set_xlim(-width / 2, len(ind))
    ax.set_xticks(ind + width / 2)
    xtick_names = ax.set_xticklabels(x)
    ax.set_xlabel('Similarity index')
    ax.set_ylabel('Frequency')
    # make the plot
    bar = ax.bar(ind, y, width, color='lightgray')
    # plot the median line
    adj_median = ((median - bins[0]) / (bins[-1] - bins[0]) * len(x))
    plt.axvline(x=adj_median, ymin=0, ymax=1, linewidth=1.5, color='r', linestyle='dashed')
    text_xpos = adj_median + 0.2 if adj_median < 6.5 else adj_median - 2.2
    text_ypos = 1.04 * max(y)
    med = round(median, 4)
    med_str = str(med) if len(str(med)) == 6 else str(med) + '0' * (6 - len(str(med)))
    ax.text(text_xpos,
            text_ypos,
            'median: {}'.format(med_str),
            color='r',
            weight='semibold')
    # save the final figure
    plt.savefig(os.path.join(output_dir, ofile))
    plt.close()


def write_data(s1, s2, median, counts, bins, similarities, ofile, args):
    bin_midpoints = [b + (bins[1] - bins[0]) / 2 for b in bins]
    ostring = 'samples: {} and {}'.format(s1, s2)
    ostring += '\n' + ('=' * len(ostring)) + '\n\n'
    ostring += 'method: {} {}\n\n'.format(args.method.title(),
                                        'divergence' if args.method == 'kullback-leibler' else 'similarity')
    ostring += 'median: {}\n\n'.format(median)
    ostring += 'iterations: {}\n'.format(len(similarities))
    ostring += 'sequences per iteration: {}\n\n'.format(args.chunksize)
    ostring += 'score bin (midpoint)\tcount\n'
    for b, c in zip(bin_midpoints, counts):
        ostring += '{}\t{}\n'.format(b, c)
    ostring += '\nraw scores:\n'
    ostring += '\n'.join([str(s) for s in similarities])
    open(ofile, 'w').write(ostring)


def write_output(s1, s2, median, counts, bins, similarities, args, control=False):
    ctrl = 'control_' if control else ''
    m = '{}'.format('divergences' if args.method == 'kullback-leibler' else 'similarities')
    fig_file = os.path.join(args.output, '{}_{}_{}_{}{}.pdf'.format(s1, s2, args.method, ctrl, m))
    data_file = os.path.join(args.output, '{}_{}_{}_{}data.txt'.format(s1, s2, args.method, ctrl))
    make_sim_plot(counts, bins, median, fig_file, args.output)
    write_data(s1, s2, median, counts, bins, similarities, data_file, args)


def update_scores(s1, s2, median, scores):
    if s1 not in scores:
        scores[s1] = {}
    scores[s1][s2] = median
    return scores




# ================================================
#
#                  PRINTING
#
# ================================================



def print_method(method):
    logger.info('METHOD: {}'.format(method.title()))


def print_collection_info(collection):
    coll_string = '{0}{1}{0}'.format(' ' * 15, collection)
    logger.info('')
    logger.info('')
    logger.info('=' * len(coll_string))
    logger.info(coll_string)
    logger.info('=' * len(coll_string))
    logger.info('')


def print_single_pair_info(p1, p2):
    logger.info('')
    logger.info('Found two collections: {} and {}'.format(p1, p2))
    logger.info('')
    logger.info('Making a single comparison:')
    logger.info('---------------------------')
    logger.info('{} vs {}'.format(p1, p2))


def print_multiple_pairs_info(pairs):
    logger.info('')
    collections = sorted(list(set([c for tup in pairs for c in tup])))
    logger.info('Found {} collections: {}, and {}'.format(len(collections),
        ', '.join(collections[:-1]), collections[-1]))
    logger.info('')
    comp_string = 'Making {} comparisons:'.format(len(pairs))
    logger.info(comp_string)
    logger.info('-' * len(comp_string))
    for p1, p2 in pairs:
        logger.info('{} vs {}'.format(p1, p2))


def print_pair_info(s1, s2):
    logger.info('')
    logger.info('')
    compare_string = 'Comparing {} and {}'.format(s1, s2)
    logger.info(compare_string)
    logger.info('-' * len(compare_string))


def print_final_results(scores, control=False):
    # '''
    # Prints nicely formatted results to sys.stdout.

    # Input
    # 2D dict, containing scores for pairs of samples
    # '''
    if not scores:
        return 0
    logger.info('')
    if control:
        logger.info('Control scores:')
    else:
        logger.info('Experimental scores:')
    for s1 in sorted(scores.keys()):
        logger.info(s1)
        logger.info('-' * len(s1))
        for s2 in sorted(scores[s1].keys()):
            logger.info('{}\t{}'.format(s2, round(scores[s1][s2], 4)))
        logger.info('')


def run(**kwargs):
    '''
    Performs repertoire-level comparison of antibody sequencing datasets.

    Currently, the only metric for comparison is V-gene usage frequency. Additional measures
    are in the works (such as comparisons based on clonality).

    Args:

        db (str): MongoDB database name.

        collection1 (str): Name of the first MongoDB collection to query for comparison.
            If both ``collection1`` and ``collection2`` are provided, ``collection1`` will
            be compared only to ``collection2``.
            If neither ``collection1`` nor ``collection2`` are provided, all collections in
            ``db`` will be processed iteratively (all pairwise comparisons will be made).
            If ``collection1`` is provided but ``collection2`` is not, ``collection1`` will
            be iteratively compared to all other collections in ``db``.

        collection2 (str): Name of the second MongoDB collection to query for comparison.
            If both ``collection1`` and ``collection2`` are provided, ``collection1`` will
            be compared only to ``collection2``.
            If neither ``collection1`` nor ``collection2`` are provided, all collections in
            ``db`` will be processed iteratively (all pairwise comparisons will be made).

        collection_prefix (str): All collections beginning with ``collection_prefix`` will
            be iteratively compared (all pairwise comparisons will be made).

        ip (str): IP address of the MongoDB server. Default is ``localhost``.

        port (int): Port of the MongoDB server. Default is ``27017``.

        user (str): Username with which to connect to the MongoDB database. If either
            of ``user`` or ``password`` is not provided, the connection to the MongoDB
            database will be attempted without authentication.

        password (str): Password with which to connect to the MongoDB database. If either
            of ``user`` or ``password`` is not provided, the connection to the MongoDB
            database will be attempted without authentication.

        chunksize (int): Number of sequences for each iteration. Default is 100,000.

        iterations (int): Number of iterations to perform on each pair of samples.
            Default is 10,000

        method (str): Similarity/divergence method to used for comparison. Default is
            ``marisita-horn``. Options are:

            - ``marisita-horn``
            - ``kullback-leibler``
            - ``jensen-shannon``
            - ``jaccard``
            - ``bray-curtis``
            - ``renkonen``
            - ``cosine``

        control_similarity (bool): If ``True``, control similarity/divergence will be
            calculated, in which each sample is also compared to itself. Default is ``False``.

        chain (str): Antibody chain to be used for comparison. Options are ``heavy``, ``kappa``
            and ``lambda``. Default is ``heavy``.
    '''

    args = Args(**kwargs)
    global logger
    logger = log.get_logger('abcompare')
    main(args)


def run_standalone(args):
    logfile = args.log if args.log else os.path.join(args.output, 'abcompare.log')
    log.setup_logging(logfile)
    global logger
    logger = log.get_logger('abcompare')
    main(args)


def main(args):
    db = mongodb.get_db(args.db, args.ip, args.port,
                        args.user, args.password)
    print_method(args.method)
    pairs = get_collection_pairs(db, args)
    index_collections(db, pairs)
    prev1 = None
    scores = {}
    cscores = {}
    for pair in pairs:
        s1, s2 = pair
        curr1 = s1
        if prev1 != curr1:
            print_collection_info(s1)
            s1_all_vgenes = get_vgenes(db, s1, args.chain)
        print_pair_info(s1, s2)
        s1_vgenes, s2_vgenes = get_vgenes(db, s2, args.chain, prev_data=s1_all_vgenes)
        logger.info('')
        logger.info('Calculating similarities...')
        median, counts, bins, similarities = calculate_similarities(s1_vgenes,
                                                                    s2_vgenes,
                                                                    args)
        write_output(s1, s2, median, counts, bins, similarities, args)
        scores = update_scores(s1, s2, median, scores)
        if args.control_similarity:
            logger.info('')
            logger.info('Calculating control similarities...')
            cmedian, ccounts, cbins, csimilarities = calculate_control_similarities(s1_vgenes,
                                                                                    s2_vgenes,
                                                                                    args)
            write_output(s1, s2, cmedian, ccounts, cbins, csimilarities, args)
            cscores = update_scores(s1, s2, cmedian, cscores)
        prev1 = s1
    print_final_results(scores)
    print_final_results(cscores, control=True)


if __name__ == '__main__':
    parser = parse_args()
    args = parser.parse_args()
    logfile = args.log if args.log else os.path.join(args.output, 'abcompare.log')
    log.setup_logging(logfile)
    logger = log.get_logger('abcompare')
    main(args)



#!/usr/bin/python
# filename: tasks.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from Bio import SeqIO

from abtools.queue.celery import celery
from abtools.alignment import global_alignment
from abtools.sequence import Sequence


@celery.task
def identity(f, standard, is_aa, debug=False):
    seqs = [s for s in SeqIO.parse(open(f, 'r'), 'fasta')]
    scores = []
    for s in seqs:
        seq_id = '_'.join(s.id.split('_')[:-1])
        germ_identity = float(s.id.split('_')[-1])
        seq = str(s.seq).upper()
        matrix = 'blosum62' if is_aa else None
        aln = global_alignment(seq,
                              str(standard.seq.upper()),
                              matrix=matrix,
                              gap_open=-15,
                              gap_extend=-2,
                              score_match=1,
                              score_mismatch=0,
                              score_gap_open=0,
                              score_gap_extend=0,
                              aa=is_aa)
        score = aln.score
        aln_length = min(len(aln.aligned_query.strip('-')), len(aln.aligned_target.strip('-')))
        norm_score = 100. * score / aln_length
        norm_score = round(norm_score, 1)
        scores.append((seq_id, norm_score, germ_identity))
    return scores



#!/usr/bin/python
# filename: celeryconfig.py


###########################################################################
#
# Copyright (c) 2014 Bryan Briney.  All rights reserved.
#
# @version: 1.0.0
# @author: Bryan Briney
# @license: MIT (http://opensource.org/licenses/MIT)
#
###########################################################################



# config file for Celery Daemon

# default RabbitMQ broker
# BROKER_URL = 'amqp://'

# Redis broker
BROKER_URL = 'redis://master:6379/0'

# default RabbitMQ backend
# CELERY_RESULT_BACKEND = 'amqp://'

# Redis backend
CELERY_RESULT_BACKEND = 'redis://master:6379/0'

# Additional Redis-specific configs
BROKER_TRANSPORT_OPTIONS = {'fanout_prefix': True,
							'fanout_patterns': True,
							'visibility_timeout': 300}

# Other configs
CELERYD_MAX_TASKS_PER_CHILD = 32
CELERY_ACKS_LATE = True



#!/usr/bin/python
# filename:  codons.py


#
# Copyright (c) 2015 Bryan Briney
# License:  The MIT license (http: //opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#



codon_lookup = {"TTT": "F", "TTC": "F", "TTA": "L", "TTG": "L",
          "TCT": "S", "TCC": "S", "TCA": "S", "TCG": "S",
          "TAT": "Y", "TAC": "Y", "TAA": "*", "TAG": "*",
          "TGT": "C", "TGC": "C", "TGA": "*", "TGG": "W",
          "CTT": "L", "CTC": "L", "CTA": "L", "CTG": "L",
          "CCT": "P", "CCC": "P", "CCA": "P", "CCG": "P",
          "CAT": "H", "CAC": "H", "CAA": "Q", "CAG": "Q",
          "CGT": "R", "CGC": "R", "CGA": "R", "CGG": "R",
          "ATT": "I", "ATC": "I", "ATA": "I", "ATG": "M",
          "ACT": "T", "ACC": "T", "ACA": "T", "ACG": "T",
          "AAT": "N", "AAC": "N", "AAA": "K", "AAG": "K",
          "AGT": "S", "AGC": "S", "AGA": "R", "AGG": "R",
          "GTT": "V", "GTC": "V", "GTA": "V", "GTG": "V",
          "GCT": "A", "GCC": "A", "GCA": "A", "GCG": "A",
          "GAT": "D", "GAC": "D", "GAA": "E", "GAG": "E",
          "GGT": "G", "GGC": "G", "GGA": "G", "GGG": "G", }



#!/usr/bin/python
# filename: _finder.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import multiprocessing as mp
import platform
import os
import subprocess as sp
import sys
import tempfile
from threading import Thread
import time

import numpy as np
import pandas as pd

from pymongo import MongoClient

from Bio import SeqIO

import matplotlib as mpl
import matplotlib.pyplot as plt

import seaborn as sns

from abtools import color, log, mongodb
from abtools.utils import progbar


def parse_args():
    import argparse
    parser = argparse.ArgumentParser("For a MongoDB collection, plots the germline divergence against the sequence identity to a given 'subject' sequence.")
    parser.add_argument('-d', '--database', dest='db', required=True,
                        help="Name of the MongoDB database to query. Required.")
    parser.add_argument('-c', '--collection', dest='collection', default=None,
                        help="Name of the MongoDB collection to query. \
                        If not provided, all collections in the given database will be processed iteratively.")
    parser.add_argument('--collection-prefix', dest='collection_prefix', default=None,
                        help="If supplied, will iteratively process only collections beginning with <collection_prefix>.")
    parser.add_argument('-o', '--output', dest='output_dir', default=None,
                        help="Output directory figure files. If not provided, figures will not be generated. \
                        Directory will be created if it does not already exist.")
    parser.add_argument('-t', '--temp', dest='temp_dir', required=True,
                        help="Directory for temporary storage. \
                        Will be created if it does not already exist. Required.")
    parser.add_argument('-l', '--log', dest='log', default=None,
                        help="The log file, to which the blast_parse log info will be written. \
                        Default is <output>/abfinder.log.")
    parser.add_argument('-C', '--cluster', dest="cluster", default=False, action='store_true',
                        help="Use if performing computation on a Celery cluster. \
                        If set, input files will be split into many subfiles and passed to a Celery queue. \
                        If not set, input files will still be split, \
                        but will be distributed to local processors using multiprocessing.")
    parser.add_argument('-i', '--ip', dest='ip', default='localhost',
                        help="The IP address for the MongoDB server.  \
                        Defaults to 'localhost'.")
    parser.add_argument('--port', dest='port', default=27017,
                        help="The port for the MongoDB server. Defaults to '27017'.")
    parser.add_argument('-u', '--user', dest='user', default=None,
                        help="Username for the MongoDB server. Not used if not provided.")
    parser.add_argument('-p', '--password', dest='password', default=None,
                        help="Password for the MongoDB server. Not used if not provided.")
    parser.add_argument('-s', '--standard', dest='standard', required=True,
                        help='Path to a file containing the standard sequence(s) for which \
                        identity/divergence will be calculated, in FASTA format. \
                        All sequences in the standard file will iteratively processed. Required')
    parser.add_argument('-q', '--chain', dest='chain', default='heavy',
                        choices=['heavy', 'kappa', 'lambda', 'light'],
                        help="The chain type of the subject sequence. \
                        Options are 'heavy', 'kappa', 'lambda' and 'light'. \
                        Default is 'heavy'.")
    parser.add_argument('-n', '--no_update', dest='update', action='store_false', default=True,
                        help="Does not update the MongoDB with AbFinder info. \
                        Can save some time if the identity calculations aren't needed again.")
    parser.add_argument('--no_figure', dest='make_figure', action='store_false', default=True,
                        help="Does not make the identity/divergence figure. \
                        Useful if you don't want the figure, just the identity info written to the database.")
    parser.add_argument('--single-process-update', dest='single_process_update', action='store_true', default=False,
                        help="Perform the MongoDB update using a single process (without multiprocessing).")
    parser.add_argument('--update-threads', dest='update_threads', type=int, default=25,
                        help="Number of threads to use when update the MongoDB database. Default is 25.")
    parser.add_argument('-N', '--nucleotide', dest='is_aa', action='store_false', default=True,
                        help="Use nucleotide sequences for alignment. Default is amino acid sequences. \
                        Ensure standard format matches.")
    parser.add_argument('-x', '--xmin', dest='x_min', type=int, default=-1,
                        help="Minimum X-axis (germline divergence) value for the AbCompare plot. Default is -1.")
    parser.add_argument('-X', '--xmax', dest='x_max', type=int, default=35,
                        help="Maximum X-axis (germline divergence) value for the AbCompare plot. Default is 35.")
    parser.add_argument('-y', '--ymin', dest='y_min', type=int, default=65,
                        help="Minimum Y-axis (mAb identity) value for the AbCompare plot. Default is 65.")
    parser.add_argument('-Y', '--ymax', dest='y_max', type=int, default=101,
                        help="Maximum Y-axis (mAb identity) value for the AbCompare plot. Default is 101.")
    parser.add_argument('-g', '--gridsize', dest='gridsize', type=int, default=0,
                        help="Gridsize for the AbFinder hexbin plot. \
                        Default is 36 for amino acid sequences and 50 for nucleotide sequences.")
    parser.add_argument('--colormap', dest='colormap', default='Blues',
                        help="Colormap to be used in the AbFinder hexbin plots. \
                        Can accept a matplotlib cmap or the name of one of matplotlib's builtin cmaps. \
                        Default is 'Blues'.")
    parser.add_argument('--mincount', dest='mincount', default=3, type=int,
                        help="Minimum number of sequences in a hexbin for that hexbin to be colored. \
                        Default is 3.")
    parser.add_argument('--skip-padding', dest='remove_padding', default=True, action='store_false',
                        help="If set, will not remove padding field from MongoDB.")
    parser.add_argument('-D', '--debug', dest="debug", action='store_true', default=False,
                        help="If set, will write all failed/exception sequences to file \
                        and should give more informative errors.")
    return parser


class Args(object):
    def __init__(self, db=None, collection=None,
                 output=None, temp=None, log=None, cluster=False,
                 ip='localhost', port=27017, user=None, password=None, update=True,
                 standard=None, chain='heavy', is_aa=True,
                 x_min=-1, x_max=35, y_min=65, y_max=101, gridsize=0, mincount=3,
                 colormap='Blues', debug=False):
        super(Args, self).__init__()
        if not all([db, output, temp, standard]):
            err = 'You must provide a MongoDB database name, output and temp directories, \
                and a file containing one or more comparison (standard) sequences in FASTA format.'
            raise RuntimeError(err)
        self.db = db
        self.collection = collection
        self.output_dir = output
        self.temp_dir = temp
        self.log = log
        self.cluster = bool(cluster)
        self.ip = ip
        self.port = int(port)
        self.user = user
        self.password = password
        self.standard = standard
        if chain not in ['heavy', 'kappa', 'lambda', 'light']:
            err = 'Please select an appropriate chain. \
                Valid choices are: heavy, light, kappa and lambda.'
            raise RuntimeError(err)
        self.chain = chain
        self.update = bool(update)
        self.is_aa = bool(is_aa)
        self.x_min = int(x_min)
        self.x_max = int(x_max)
        self.y_min = int(y_min)
        self.y_max = int(y_max)
        self.gridsize = int(gridsize)
        mincount = int(mincount)
        self.colormap = colormap
        self.debug = bool(debug)



# ================================================
#
#             FILES AND DIRECTORIES
#
# ================================================



def make_directories(args):
    for d in [args.output_dir, args.temp_dir]:
        if d:
            _make_direc(d, args)


def _make_direc(d, args):
    if not os.path.exists(d):
        os.makedirs(d)
    if args.cluster:
        cmd = 'sudo chmod 777 {}'.format(d)
        p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
        stdout, stderr = p.communicate()


def get_standards(args):
    standards = []
    for s in SeqIO.parse(open(args.standard, 'r'), 'fasta'):
        standards.append(s)
    return standards


def get_chain(args):
    if args.chain == 'light':
        return ['kappa', 'lambda']
    return [args.chain, ]


def get_sequences(db, collection, temp_dir, args):
    files = []
    fastas = []
    chunksize = 1000
    seq_counter = 0
    total_seq_counter = 0
    query_results = query(db, collection, args)
    iden_field = 'aa_identity' if args.is_aa else 'nt_identity'
    vdj_field = 'vdj_aa' if args.is_aa else 'vdj_nt'
    for seq in query_results:
        fastas.append('>{}_{}\n{}'.format(seq['seq_id'], seq[iden_field]['v'], seq[vdj_field]))
        seq_counter += 1
        total_seq_counter += 1
        if seq_counter == chunksize:
            files.append(write_to_temp_file(fastas, temp_dir))
            fastas = []
            seq_counter = 0
    if fastas:
        files.append(write_to_temp_file(fastas, temp_dir))
    return files


def write_to_temp_file(fastas, temp_dir):
    tfile = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
    tfile.write('\n'.join(fastas))
    tfile.close()
    return tfile.name


def clean_up(files):
    for f in files:
        os.unlink(f)



# ================================================
#
#                   MONGO
#
# ================================================



def query(db, collection, args):
    coll = db[collection]
    chain = get_chain(args)
    mongodb.index(db, collection, ['chain'])
    print_query_info()
    iden_field = 'aa_identity.v' if args.is_aa else 'nt_identity.v'
    vdj_field = 'vdj_aa' if args.is_aa else 'vdj_nt'
    return coll.find({'chain': {'$in': chain}, 'prod': 'yes'}, {'_id': 0, 'seq_id': 1, iden_field: 1, vdj_field: 1})


def chunker(l, n):
    'Generator that produces n-length chunks from iterable l.'
    for i in xrange(0, len(l), n):
        yield l[i:i + n]


def update_db(db, standard, scores, collection, args):
    db = mongodb.get_db(args.db, args.ip, args.port, args.user, args.password)
    print_index_info()
    mongodb.index(db, collection, ['seq_id'])
    print_update_info()
    start = time.time()
    conn = mongodb.get_connection(args.ip, args.port,
        args.user, args.password)
    mongo_version = conn.server_info()['version']
    standard = standard.replace('.', '_')
    g = scores.groupby('identity')
    groups = regroup(g.groups)


    for g in range(0, len(groups), args.update_threads):
        tlist = []
        for group in groups[g:g + args.update_threads]:
            t = Thread(target=update, args=(db, collection, group, standard, mongo_version, args))
            t.start()
            tlist.append(t)
        for t in tlist:
            t.join()
        progbar.progress_bar(g + args.update_threads, len(groups))


    # if platform.system().lower() == 'darwin' or args.debug or args.single_process_update:
    #     for i, group in enumerate(groups):
    #         update(db, collection, group, standard, mongo_version, args)
    #         progbar.progress_bar(i, len(groups))
    # else:
    #     p = mp.Pool(processes=25)
    #     async_results = []
    #     for group in groups:
    #         async_results.append(p.apply_async(update, args=(db, collection, group, standard, mongo_version, args)))
    #     monitor_update(async_results)
    #     p.close()
    #     p.join()
    print('')
    run_time = time.time() - start
    logger.info('Updating took {} seconds. ({} sequences per second)'.format(round(run_time, 2),
        round(len(scores) / run_time, 1)))


def update(db, collection, data, standard, version, args):
    db = mongodb.get_db(args.db, args.ip, args.port, args.user, args.password)
    coll = db[collection]
    score = data[0]
    ids = data[1]
    mab_id_field = 'mab_identity_aa' if args.is_aa else 'mab_identity_nt'
    if int(version.split('.')[0]) < 3:
        result = coll.update({'seq_id': {'$in': ids}},
                    {'$set': {'{}.{}'.format(mab_id_field, standard.lower()): float(score)}},
                    multi=True)
    else:
        result = coll.update_many({'seq_id': {'$in': ids}},
                         {'$set': {'{}.{}'.format(mab_id_field, standard.lower()): float(score)}})

        if args.debug:
            print('matched: {}'.format(result.matched_count))
            print('modified: {}'.format(result.modified_count))


def monitor_update(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        finished = len([r for r in results if r.ready()])
        progbar.progress_bar(finished, jobs)
    progbar.progress_bar(finished, jobs)


def regroup(oldgs):
    newgs = []
    for og in oldgs:
        if len(oldgs[og]) <= 500:
            newgs.append((og, oldgs[og]))
        else:
            for ng in chunker(oldgs[og], 500):
                newgs.append((og, ng))
    return newgs




# ================================================
#
#                  FIGURES
#
# ================================================



def make_figure(standard_id, scores, collection, args):
    print_fig_info()
    sns.set_style('white')
    fig_file = os.path.join(args.output_dir, '{0}_{1}_{2}.pdf'.format(args.db, collection, standard_id))
    x = list(scores['germ_divergence'].values)
    y = list(scores['identity'].values)
    xy_vals = zip(x, y)
    trunc_xy_vals = [v for v in xy_vals if v[0] <= args.x_max and v[1] >= args.y_min]
    x = [v[0] for v in trunc_xy_vals]
    y = [v[1] for v in trunc_xy_vals]
    # To make sure the gridsize is correct (since it's based on the actual values)
    # I need to add a single value near the max and min of each axis.
    # They're added just outside the visible plot, so there's no effect on the plot.
    x.extend([args.x_min - 1, args.x_max + 1])
    y.extend([args.y_min - 1, args.y_max + 1])
    # plot params
    cmap = color.get_cmap(args.colormap)
    plt.subplots_adjust(hspace=0.95)
    plt.subplot(111)
    plt.hexbin(x, y, bins='log', cmap=cmap, mincnt=3, gridsize=set_gridsize(args))
    plt.title(standard_id, fontsize=18)
    # set and label axes
    plt.axis([args.x_min, args.x_max, args.y_min, args.y_max])
    plt.xlabel('Germline divergence')
    plt.ylabel('{0} identity'.format(standard_id))
    # make and label the colorbar
    cb = plt.colorbar()
    cb.set_label('Sequence count (log10)', labelpad=10)
    # save figure and close
    plt.savefig(fig_file)
    plt.close()


def set_gridsize(args):
    if args.gridsize:
        return args.gridsize
    elif args.is_aa:
        return 36
    return 50




# ================================================
#
#                  PRINTING
#
# ================================================



def print_abfinder_start():
    logger.info('')
    logger.info('')
    logger.info('')
    logger.info('-' * 25)
    logger.info('ABFINDER')
    logger.info('-' * 25)


def print_standards_info(standards):
    logger.info('')
    logger.info('Found {} standard sequence(s):'.format(len(standards)))
    logger.info(', '.join([s.id for s in standards]))


def print_collections_info(collections):
    logger.info('')
    logger.info('Found {} collection(s):'.format(len(collections)))
    logger.info(', '.join(collections))


def print_single_standard(standard):
    standard_id_string = '{}'.format(standard.id)
    logger.info('')
    logger.info(standard_id_string)
    logger.info('-' * len(standard_id_string))


def print_single_collection(collection):
    logger.info('')
    logger.info('')
    logger.info(collection)
    logger.info('-' * len(collection))


def print_query_info():
    logger.info('Querying for comparison sequences...')


def print_remove_padding():
    logger.info('')
    logger.info('Removing MongoDB padding...')


def print_fig_info():
    logger.info('Making identity/divergence figure...')


def print_index_info():
    logger.info('Indexing the MongoDB collection...')


def print_update_info():
    logger.info('Updating the MongoDB database with identity scores:')




# ================================================
#
#                IDENTITY JOBS
#
# ================================================



def run_jobs(files, standard, args):
    logger.info('Running AbCompare...')
    if args.cluster:
        return _run_jobs_via_celery(files, standard, args)
    else:
        return _run_jobs_via_multiprocessing(files, standard, args)


def _run_jobs_via_multiprocessing(files, standard, args):
    from abtools.queue.tasks import identity
    results = []
    if args.debug:
        for f in files:
            results.extend(identity(f, standard, args.is_aa, args.debug))
    else:
        p = mp.Pool()
        async_results = []
        for f in files:
            async_results.append(p.apply_async(identity, (f, standard, args.is_aa)))
        monitor_mp_jobs(async_results)
        for a in async_results:
            results.extend(a.get())
        p.close()
        p.join()
    ids = [r[0] for r in results]
    identities = pd.Series([r[1] for r in results], index=ids)
    divergences = pd.Series([100. - r[2] for r in results], index=ids)
    d = {'identity': identities, 'germ_divergence': divergences}
    df = pd.DataFrame(d)
    return df


def monitor_mp_jobs(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        ready = [ar for ar in results if ar.ready()]
        finished = len(ready)
        update_progress(finished, jobs)
    print('')


def _run_jobs_via_celery(files, standard, args):
    from abtools.queue.tasks import identity
    async_results = []
    for f in files:
        async_results.append(identity.delay(f, standard, args.is_aa))
    succeeded, failed = monitor_celery_jobs(async_results)
    scores = []
    for s in succeeded:
        scores.extend(s.get())
    ids = [r[0] for r in scores]
    identities = pd.Series([r[1] for r in scores], index=ids)
    divergences = pd.Series([r[2] for r in scores], index=ids)
    d = {'identity': identities, 'germ_divergence': divergences}
    df = pd.DataFrame(d)
    return df


def monitor_celery_jobs(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        succeeded = [ar for ar in results if ar.successful()]
        failed = [ar for ar in results if ar.failed()]
        finished = len(succeeded) + len(failed)
        update_progress(finished, jobs, failed=len(failed))
    print('')
    return succeeded, failed


def update_progress(finished, jobs):
    pct = int(100. * finished / jobs)
    ticks = pct / 2
    spaces = 50 - ticks
    prog_bar = '\r({}/{}) |{}{}|  {}%'.format(finished, jobs, '|' * ticks, ' ' * spaces, pct)
    sys.stdout.write(prog_bar)
    sys.stdout.flush()


def run(**kwargs):
    '''
    Mines NGS datasets for identity to known antibody sequences.

    All of ``db``, ``output``, ``temp`` and ``standard`` are required.


    Args:

        db (str): Name of a MongoDB database to query.

        collection (str): Name of a MongoDB collection. If not provided, all collections
            in ``db`` will be processed iteratively.

        output_dir (str): Path to the output directory, into which identity/divergence
            figures will be deposited.

        temp_dir (str): Path to a temporary directory.

        log (str): Path to a log file. If not provided, log information will not be retained.

        ip (str): IP address of the MongoDB server. Default is ``localhost``.

        port (str): Port of the MongoDB server. Default is ``27017``.

        user (str): Username with which to connect to the MongoDB database. If either
            of ``user`` or ``password`` is not provided, the connection to the MongoDB
            database will be attempted without authentication.

        password (str): Password with which to connect to the MongoDB database. If either
            of ``user`` or ``password`` is not provided, the connection to the MongoDB
            database will be attempted without authentication.

        standard (path): Path to a FASTA-formatted file containing one or more 'standard'
            sequences, against which the NGS sequences will be compared.

        chain (str): Antibody chain. Choices are 'heavy', 'kappa', 'lambda', and 'light'.
            Default is 'heavy'. Only NGS sequences matching ``chain`` (with 'light' covering
            both 'kappa' and 'lambda') will be compared to the ``standard`` sequences.

        update (bool): If ``True``, the MongoDB record for each NGS sequence will be updated
            with identity information for each standard. If ``False``, the updated is skipped.
            Default is ``True``.

        is_aa (bool): If ``True``, the ``standard`` sequences are amino acid sequences. If
            ``False``, they are nucleotide seqeunces. Default is ``False``.

        x_min (int): Minimum x-axis value on identity/divergence plots.

        x_max (int): Maximum x-axis value on identity/divergence plots.

        y_min (int): Minimum y-axis value on identity/divergence plots.

        y_max (int): Maximum y-axis value on identity/divergence plots.

        gridsize (int): Relative size of hexbin grids.

        mincount (int): Minimum number of sequences in a hexbin for the bin to be colored.
            Default is 3.

        colormap (str, colormap): Colormap to be used for identity/divergence plots.
            Default is ``Blues``.

        debug (bool): If ``True``, more verbose logging.
   '''
    args = Args(**kwargs)
    global logger
    logger = log.get_logger('abfinder')
    main(args)


def run_standalone(args):
    logfile = args.log if args.log else os.path.join(args.output_dir, 'abfinder.log')
    log.setup_logging(logfile)
    global logger
    logger = log.get_logger('abfinder')
    main(args)


def main(args):
    print_abfinder_start()
    db = mongodb.get_db(args.db, args.ip, args.port,
        args.user, args.password)
    make_directories(args)
    standards = get_standards(args)
    print_standards_info(standards)
    collections = mongodb.get_collections(db, args.collection, prefix=args.collection_prefix)
    print_collections_info(collections)
    for collection in collections:
        indexed = False
        print_single_collection(collection)
        if args.remove_padding:
            print_remove_padding()
            mongodb.remove_padding(db, collection)
        seq_files = get_sequences(db, collection, args.temp_dir, args)
        for standard in standards:
            print_single_standard(standard)
            scores = run_jobs(seq_files, standard, args)
            if args.output_dir:
                make_figure(standard.id, scores, collection, args)
            if args.update:
                if not indexed:
                    mongodb.index(db, collection, 'seq_id')
                    indexed = True
                update_db(db, standard.id, scores, collection, args)
        clean_up(seq_files)


if __name__ == '__main__':
    parser = parse_args()
    args = parser.parse_args()
    logfile = args.log if args.log else os.path.join(args.output_dir, 'abfinder.log')
    log.setup_logging(logfile)
    logger = log.get_logger('abfinder')
    main(args)



#!/usr/bin/python
# filename: _phylogeny.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import subprocess as sp
import sys

import matplotlib as mpl
import seaborn as sns

from Bio import SeqIO

from abtools.phylogeny.utils import msa, tree
from abtools.phylogeny.utils.timepoint import Timepoint
from abtools.sequence import Sequence



def parse_args():
    import argparse
    parser = argparse.ArgumentParser("")
    parser.add_argument('-i', '--input', dest='input', required=True,
                        help="Input file, containing sequences in FASTA format with the timepoint appended to the sequence ID. \
                        Required.")
    parser.add_argument('-o', '--output', dest='output', required=True,
                        help="Output directory for the FASTA, alignment, tree and figure files. \
                        Required")
    parser.add_argument('-r', '--root', dest='root', required=True,
                        help="FASTA file containing a single sequence that will be used to root the phylogenetic tree. \
                        Required.")
    parser.add_argument('-m', '--mabs', dest='mabs', default=None,
                        help="FASTA file containing sequences of monoclonal antibodies of the appropriate chain.")
    parser.add_argument('-a', '--alignment', dest='alignment', default=None,
                        help="Provide a pre-aligned dataset. \
                        Ensure that the root sequence is labeled as 'root' and that all mAb sequences are pre-pended with 'mab_'. \
                        If --alignment is provided, -i, -r and -m are ignored.")
    parser.add_argument('-n', '--newick', dest='newick', default=None,
                        help="Provide a pre-generated Newick tree file. \
                        Ensure that the root sequence is labeled as 'root' and that all mAb sequences are pre-pended with 'mab_'. \
                        If --newick is provided -i, -r, -m and -a are ignored.")
    parser.add_argument('-s', '--sample_id', dest='sample_id', default=None,
                        help="Sample ID. If not provided, the sample ID will be inferred from the input file.")
    parser.add_argument('-t', '--timepoints', dest='timepoints', default=None,
                        help="Tab-delimited file, of the following format (one per line): \
                        TimepointName    TimepointOrder    TimepointColor\
                        TimepointName is the name appended to the sequences in the input file.\
                        TimepointOrder is an integer that indicates the order in which the timepoints should be sorted.\
                        TimepointColor is an RGB or hex value that will be used to color the phylogenetic tree.\
                        If mAb sequences are provided, the 'mab' entry will be used to sort/color the mAb sequences.\
                        If not provided, colors will be automatically selected and timepoints will be determined by a simple \
                        sort of the raw timepoint values parsed from the input file.")
    parser.add_argument('-A', '--aa', dest='is_aa', default=False, action='store_true',
                        help="If used, all input files contain amino acid sequences. \
                        Default is nucleotide sequences.")
    parser.add_argument('-D', '--delimiter', dest='delimiter', default='_',
                        help="Delimiter that separates the timepoint and sequence ID. \
                        Cannot use ':', ';' or '=', since these can screw up the tree file. \
                        Default is '_'.")
    parser.add_argument('-S', '--scale', dest='scale', default=None,
                        help="Scale for the resulting ete2 tree. \
                        If not provided, the ete2 default value will be used.")
    parser.add_argument('--label-nodes', dest='label_nodes', default='mab',
                        help="Type of nodes to be labeled. \
                        Options are: all, none, no-root, mab, input, and root.")
    parser.add_argument('--label-fontsize', dest='label_fontsize', default=12,
                        help="Fontsize for node lables. Default is 12.")
    parser.add_argument('-b', '--branch-vertical-margin', dest='branch_vertical_margin', default=None,
                        help="Branch vertical margin for the resulting ete2 tree. \
                        If not provided, the ete2 default value will be used..")
    parser.add_argument('--sequence-key', dest='sequence_key', default='vdj_nt',
                        help="If providing a list of sequence dicts, the dict key to use \
                        as the sequence for alignment and phylogeny. \
                        Default is 'vdj_nt'.")
    parser.add_argument('--name-key', dest='name_key', default='seq_id',
                        help="If providing a list of sequence dicts, the dict key to use \
                        as the sequence name for alignment and phylogeny. \
                        Default is 'seq_id'.")
    parser.add_argument('--orientation', dest='tree_orientation', default=0, type=int,
                        help="Tree orientation. If 0, tree will be drawn from left to right. \
                        if 1, tree will be drawn from right to left. \
                        Default is 0.")
    return parser


class Args(object):
    def __init__(self, input=None, output=None,
                 root=None, mabs=None,
                 alignment=None, newick=None,
                 sample_id=None, is_aa=False,
                 timepoints=None, delimiter='_',
                 label_fontsize=12, label_nodes='mab',
                 scale=None, branch_vertical_margin=None, tree_orientation=0,
                 sequence_key='vdj_nt', name_key='seq_id'):
        super(Args, self).__init__()
        if not any([input, alignment, newick]):
            print('An input must be provided, either <input>, <alignment> or <newick>.')
            sys.exit(1)
        if output is None:
            print('An output file must be provided.')
            sys.exit(1)
        self.input = input
        self.output = output
        self.root = root
        self.mabs = mabs
        self.alignment = alignment
        self.newick = newick
        self.sample_id = sample_id
        self.timepoints = timepoints
        self.is_aa = is_aa
        self.delimiter = delimiter
        self.scale = scale
        self.branch_vertical_margin = branch_vertical_margin
        self.label_nodes = label_nodes
        self.label_fontsize = label_fontsize
        self.sequence_key = sequence_key
        self.name_key = name_key
        self.tree_orientation = int(tree_orientation)



# ================================================
#
#                 INPUT PARSING
#
# ================================================



def parse_seqs(args):
    seqs = parse_input_file(args.input, args)
    if args.root:
        seqs += parse_root(args.root, args)
    if args.mabs:
        seqs += parse_mabs(args.mabs, args.delimiter, args)
    timepoints = list(set([s.id.split(args.delimiter)[0] for s in seqs]))
    return seqs, timepoints


def parse_input_file(input, args):
    if type(input) in [list, tuple]:
        seqs = [Sequence(s, id_key=args.name_key, seq_key=args.sequence_key) for s in input]
    else:
        seqs = [Sequence(s.id, str(s.seq)) for s in SeqIO.parse(open(input, 'r'), 'fasta')]
    return seqs


def parse_root(root, args):
    if type(root) == dict:
        _root = Sequence(root, id='root', seq_key=args.sequence_key)
    elif type(root) in [list, tuple]:
        _root = Sequence(root)
    elif type(root) in [str, unicode] and not os.path.isfile(root):
        _root = Sequence(root, id='root')
    else:
        r = SeqIO.read(open(root_file, 'r'), 'fasta')
        _root = Sequence('root', str(r.seq))
    return [_root, ]


def parse_mabs(mabs, delimiter, args):
    if type(mabs) in [list, tuple]:
        seqs = [Sequence(m, id_key=args.name_key, seq_key=args.sequence_key) for m in mabs]
    elif type(mabs) == dict:
        seqs = [Sequence(mabs), ]
    else:
        seqs = [Sequence(s.id, str(s.seq)) for s in SeqIO.parse(open(mabs_file, 'r'), 'fasta')]
    for seq in seqs:
        if seq.id.split(delimiter)[0] != 'mab':
            seq.id = 'mab{}{}'.format(delimiter, seq.id)
    return seqs


def parse_timepoints(tps, args):
    if type(args.timepoints) in [list, tuple]:
        timepoints = [Timepoint(*t) for t in args.timepoints]
    elif args.timepoints is not None:
        timepoints = []
        with open(args.timepoints, 'r') as f:
            for line in f:
                name, order, color = line.strip().split('\t')
                if name in tps:
                    timepoints.append(Timepoint(name, order, color))
    else:
        colors = sns.hls_palette(len(tps), l=0.5, s=0.9)
        timepoints.append(Timepoint('root', 0, colors[0]))
        for i, tp in enumerate(sorted([t for t in tps if t != 'root'])):
            timepoints.append(Timepoint(tp, i + 1, colors[i + 1]))
    return timepoints


def make_msa(seqs, args):
    if args.sample_id:
        sample = args.sample_id
    else:
        sample = '.'.join(os.path.basename(args.input).split('.')[:-1])
    fasta_file = os.path.join(args.output, '{}.fasta'.format(sample))
    alignment = msa.align(seqs, fasta_file)
    return alignment


def make_tree(alignment, timepoints, args):
    return tree.make_tree(alignment,
                          timepoints,
                          args.delimiter,
                          args.is_aa,
                          args.scale,
                          args.branch_vertical_margin,
                          args.label_fontsize,
                          args.label_nodes,
                          args.tree_orientation)


def run(**kwargs):
    '''
    Builds a phylogenetic representation of antibody sequences.

    ``output`` is required, as well as one of ``input``, ``alignment`` or ``newick``.

    Args:

        input (str): Can be one of three things:

            1. Path to a FASTA-formatted file containing input sequences.
            2. A list of AbTools ``Sequence`` objects.
            3. A list of dictionaries, containing at minimum ``name_key`` and ``seq_key``.

        output (str): Path to the output directory, into which tree images and
            all intermediate files will be deposited.

        root (str): Path to a FASTA-formatted file containing a single sequence
            which will be used to root the tree. If not provided, tree will be unrooted.

        mabs (str): Path to a FASTA-formatted file containing mAb sequences. If supplying
            both mAb sequences and NGS sequences, passing the mAb sequences separately
            allows you to modify their representation separately (for example, show sequence
            IDs for just the mAb sequences).

        alignment (str): Path to a multiple sequence alignment, in FASTA format. If sequences
            are already aligned, this will save some computational time since the alignment
            will not be redone.

        newick (str): Path to a tree file, in Newick format. As with ``alignment``, this is
            primarily to save computational time if the tree file has already been generated.

        name_key (str): If ``input`` is a list of Sequence objects or dicts, this key will be
            used to find the sequence ID. Default is ``seq_id``.

        sequence_key (str): If ``input`` is a list of Sequence objects or dicts, this key will be
            used to find the sequence. Default is ``vdj_nt``.

        timepoints (str): Path to a Tab-delimited file, of the following format (one per line)::

                TimepointName    TimepointOrder    TimepointColor

            ``TimepointName`` should prepended to the sequences in the input file (separated by ``delimiter``).

            ``TimepointOrder`` is an integer that indicates the order in which the timepoints should be sorted.

            ``TimepointColor`` is a hex value that will be used to color the phylogenetic tree.
            If mAb sequences are provided, the 'mab' ``TimepointName`` will be used to sort/color the mAb sequences.
            If not provided, colors will be automatically selected and timepoints will be determined by a simple
            sort of the raw timepoint values parsed from the input file.

        is_aa (bool): If ``True``, input sequences will be assumed to be amino acid sequences.
            Default is ``False``, which assumes nucleotide sequences.

        delimiter (str): The delimiter used in sequence IDs to separate the timepoint from
            the sequence name. Default is ``_``.

        scale (int): Horizontal scale of the phylogeny. Default is ``None``, which uses the
            default ``ete2`` value.

        branch_vertical_margin (float): Vertical scale of the phylogeny. Default is ``None``,
            which uses the default ``ete2`` value.

        label_nodes (str): Type of nodes to be labeled. Options are: ``all``, ``none``,
            ``no-root``, ``mab``, ``input``, and ``root``.

        label_fontsize (float): Font size for the node labels.

        tree_orientation (int): If ``0``, tree is drawn from left to right. If ``1``, tree
            will be drawn from right to left (mirror). Default is ``0``.
    '''
    args = Args(**kwargs)
    main(args)


def run_standalone(args):
    logfile = args.log if args.log else os.path.join(args.output, 'abphylogeny.log')
    log.setup_logging(logfile)
    global logger
    logger = log.get_logger('abphylogeny')
    main(args)


def main(args):
    seqs, tps = parse_seqs(args)
    timepoints = parse_timepoints(tps, args)
    alignment = make_msa(seqs, args)
    make_tree(alignment, timepoints, args)


if __name__ == '__main__':
    parser = parse_args()
    args = parser.parse_args
    main(args)



#!/usr/bin/env python
# filename: _stats.py


###########################################################################
#
# Copyright (c) 2015 Bryan Briney.  All rights reserved.
#
# @version: 1.0.0
# @author: Bryan Briney
# @license: MIT (http://opensource.org/licenses/MIT)
#
###########################################################################


from __future__ import print_function

import os
import sys
import uuid
import time
import shelve
import urllib
import sqlite3
import tempfile
import argparse
import subprocess as sp
import multiprocessing as mp
from StringIO import StringIO
from collections import Counter

import numpy as np
import pandas as pd

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

from pymongo import MongoClient

from abtools import log, mongodb

from abstar.core.germline import get_germlines



def parse_args():
    parser = argparse.ArgumentParser("Computes and plots basic repertoire information from one or more antibody NGS datasets.")
    parser.add_argument('-o', '--output', dest='output', required=True,
                        help="Output directory for figure and data files. Required.")
    parser.add_argument('-t', '--temp', dest='temp_dir', required=True,
                        help="Directory for temporary files. Required.")
    parser.add_argument('-l', '--log', dest='log',
                        help="Location for the AbStats log file. \
                        If not provided, will be written to <output>/abstats.log.")
    parser.add_argument('-d', '--database', dest='db', required=True,
                        help="Name of the MongoDB database to query. Required.")
    parser.add_argument('-c', '--collection', dest='collection', default=None,
                        help="Name of the MongoDB collection to query. \
                        If not provided, all collections in the given database will be processed iteratively.")
    parser.add_argument('-i', '--ip', dest='ip', default='localhost',
                        help="IP address for the MongoDB server.  Defaults to 'localhost'.")
    parser.add_argument('--port', dest='port', default=27017, type=int,
                        help="Port for the MongoDB server.  Defaults to 27017.")
    parser.add_argument('-u', '--user', dest='user', default=None,
                        help="Username for the MongoDB server. Not used if not provided.")
    parser.add_argument('-p', '--password', dest='password', default=None,
                        help="Password for the MongoDB server. Not used if not provided.")
    parser.add_argument('-V', '--var', dest='var_plot', default=None,
                        help="Plot the distribution of variable families or genes. \
                        Options are 'fam', 'gene', or 'both'. \
                        If not provided, no plot will be made.")
    parser.add_argument('-D', '--div', dest='div_plot', default=None,
                        help="Plot the distribution of diversity families or genes. \
                        Options are 'fam', 'gene', or 'both'. \
                        If not provided, no plot will be made.")
    parser.add_argument('-J', '--join', dest='join_plot', default=None,
                        help="Plot the distribution of joining families or genes. \
                        Only option is 'gene'. \
                        If not provided, no plot will be made.")
    parser.add_argument('-3', '--cdr3', dest='cdr3_plot', default=None,
                        help="Plot the distribution of CDR3 lengths, as nucleotide or amino acid. \
                        Options are 'nt', 'aa', or 'both'. \
                        If not provided, no plot will be made.")
    parser.add_argument('-H', '--heatmap', dest='heatmap', default=False, action='store_true',
                        help="If set, generates a heatmap (or quiltplot, if you're a novice) of combined variable and joining gene use. \
                        If not provided, no plot will be made.")
    parser.add_argument('-I', '--isotype', dest='isotype_plot', default=False, action='store_true',
                        help="Plot the isotype frequency. \
                        Requires that isotypes have been identified using AbStar, or the 'isotype' field is present in the given MongoDB collection. \
                        If not provided, no plot will be made.")
    parser.add_argument('-C', '--chain', dest='chain', default='heavy', choices=['heavy', 'kappa', 'lambda'],
                        help="Select the antibody chain to analyze. \
                        Options are 'heavy', 'kappa', and 'lambda'. \
                        Default is 'heavy'.")
    parser.add_argument('-s', '--species', dest='species', default='human', choices=['human', 'mouse', 'macaque'],
                        help="Species of the sequence data. Choices are 'human', 'mouse', and 'macaque'. \
                        Default is 'human'.")
    parser.add_argument('--debug', dest='debug', action='store_true', default=False,
                        help="If set, will run in debug mode.")
    return parser



class Args(object):
    """docstring for Args"""
    def __init__(self, output=None, temp=None, log=None,
        db=None, collection=None, ip='localhost', port=27017,
        user=None, password=None,
        var_plot=None, div_plot=None, join_plot=None,
        cdr3_plot=None, heatmap=False, isotype_plot=False, chain='heavy',
        debug=False):
        super(Args, self).__init__()

        if not all([output, temp, db]):
            print('\nERROR: --output, --temp and --database are all required options.\n')
            sys.exit(1)

        self.output = output
        self.temp = temp
        self.log = log
        self.db = db
        self.collection = collection
        self.ip = ip
        self.port = int(port)
        self.user = user
        self.password = password
        self.var_plot = var_plot
        self.div_plot = div_plot
        self.join_plot = join_plot
        self.cdr3_plot = cdr3_plot
        self.heatmap = heatmap
        self.isotype_plot = isotype_plot
        self.chain = chain
        self.debug = debug



def get_db(db, ip='localhost', port=27017, user=None, password=None):
    if all([user is not None, password is not None]):
        pwd = urllib.quote_plus(password)
        uri = 'mongodb://{}:{}@{}:{}'.format(user, pwd, ip, port)
        conn = MongoClient(uri)
    else:
        conn = MongoClient(ip, port)
    return conn[db]


def get_collections(db):
    if args.collection:
        return [args.collection, ]
    collections = db.collection_names(include_system_collections=False)
    return sorted(collections)



def query(db, collection, chain):
    print('Getting sequences from MongoDB...', end='')
    sys.stdout.flush()
    c = db[collection]
    results = c.find({'chain': chain, 'prod': 'yes'},
                     {'_id': 0,
                      'v_gene.gene': 1, 'd_gene.gene': 1, 'j_gene.gene': 1,
                      'v_gene.fam': 1, 'd_gene.fam': 1, 'j_gene.fam': 1,
                      'cdr3_len': 1, 'isotype': 1})
    seqs = [r for r in results if '/OR' not in r['v_gene']['gene']]
    print('Done.\nFound {} sequences\n'.format(len(seqs)))
    return seqs


def aggregate(data, norm=True, sort_by='value', keys=None):
    '''
    Counts the number of occurances of each item in 'data'.

    Inputs
    data: a list of values.
    norm: normalize the resulting counts (as percent)
    sort_by: how to sort the retured data. Options are 'value' and 'count'.

    Output
    a non-redundant list of values (from 'data') and a list of counts.
    '''
    if keys:
        vdict = {k: 0 for k in keys}
        for d in data:
            if d in keys:
                vdict[d] += 1
    else:
        vdict = {}
        for d in data:
            vdict[d] = vdict[d] + 1 if d in vdict else 1
    vals = [(k, v) for k, v in vdict.iteritems()]
    if sort_by == 'value':
        vals.sort(key=lambda x: x[0])
    else:
        vals.sort(key=lambda x: x[1])
    xs = [v[0] for v in vals]
    if norm:
        raw_y = [v[1] for v in vals]
        total_y = sum(raw_y)
        ys = [100. * y / total_y for y in raw_y]
    else:
        ys = [v[1] for v in vals]
    return xs, ys




def cdr3_plot(seqs, collection, make_plot, chain, output_dir):
    if not make_plot:
        return None
    max_len = 40 if chain == 'heavy' else 20
    cdr3s = [s['cdr3_len'] for s in seqs if s['cdr3_len'] > 0 and s['cdr3_len'] <= max_len]
    x, y = aggregate(cdr3s, keys=range(1, max_len + 1))
    color = sns.hls_palette(7)[4]
    plot_file = os.path.join(output_dir, '{}_{}_cdr3_lengths.pdf'.format(collection, chain))
    x_title = 'CDR3 Length (AA)'
    y_title = 'Frequency (%)'
    size = (9, 4) if chain == 'heavy' else (6, 4)
    make_barplot([str(i) for i in x], y,
                 color,
                 plot_file,
                 xlabel=x_title,
                 ylabel=y_title,
                 grid=True,
                 size=size,
                 xfontsize=7)


def germline_plot(seqs, gene, collection, output_dir, level, species, chain):
    # from abtools.utils.germlines import germlines
    germs = get_germlines(species, gene, chain=chain)
    if not level:
        return None
    level = ['fam', 'gene'] if level == 'both' else [level, ]
    for l in level:
        if l == 'fam':
            keys = list(set([g.name.split('-')[0] for g in germs]))
            size = (6, 4)
        else:
            keys = list(set([g.name.split('*')[0] for g in germs]))
            if gene == 'J':
                size = (6, 4)
            elif gene == 'D':
                size = (8, 4)
            else:
                size = (10, 4)
        gtype = '{}_gene'.format(gene.lower())
        short_chain = chain[0].upper()
        data = [s[gtype][l].rstrip('D') for s in seqs if gtype in s]
        x, y = aggregate(data, keys=keys)
        x = [i.replace('IG{}{}'.format(short_chain, gene), '{}{}'.format(gene, short_chain)) for i in x]
        plot_file = os.path.join(output_dir, '{}_{}{}_{}.pdf'.format(collection, gene, chain, l))
        colors = get_germline_plot_colors(x, l)
        make_barplot(x, y,
                     colors,
                     plot_file,
                     ylabel='Frequency (%)',
                     l=l,
                     size=size)


def get_germline_plot_colors(data, l):
    fams = [d.split('-')[0].replace('/OR', '') for d in data]
    nr_fams = sorted(list(set(fams)))
    num_colors = len(nr_fams)
    rgbs = sns.hls_palette(num_colors)
    rgb_dict = {i[0]: i[1] for i in zip(nr_fams, rgbs)}
    return [rgb_dict[f] for f in fams]



def isotypes():
    pass


def vj_heatmap(seqs, collection, make_plot, species, chain, output_dir):
    if not make_plot:
        return None
    plot_file = os.path.join(output_dir, '{}_VJheatmap_{}.pdf'.format(collection, chain))
    vj_data = group_by_vj(seqs, species, chain)
    vj_df = pd.DataFrame(vj_data)
    make_heatmap(vj_df.transpose(), plot_file)


def group_by_vj(data, species, chain):
    # from abtools.utils.germlines import germlines
    vs = list(set([g.name.split('*')[0] for g in get_germlines(species, 'V', chain=chain)]))
    js = list(set([g.name.split('*')[0] for g in get_germlines(species, 'J', chain=chain)]))
    vj = {}
    for v in vs:
        vj[v] = {}
        for j in js:
            vj[v][j] = 0
    for d in data:
        v = d['v_gene']['gene']
        j = d['j_gene']['gene'].rstrip('P')
        if v[:3] != j[:3]:
            continue
        if v not in vj:
            continue
        if j not in vj[v]:
            continue
        vj[v][j] = vj[v][j] + 1 if j in vj[v] else 1
    total = len(data)
    for v in vj.keys():
        for j in vj[v].keys():
            vj[v][j] = 100. * vj[v][j] / total
    return vj






def make_barplot(x, y, colors, ofile, xlabel=None, ylabel=None, l=None, grid=False, size=None, xfontsize=None):
    sns.set_style('ticks')
    # set bar locations and width
    ind = np.arange(len(x))
    width = 0.75
    # plot objects
    if size:
        fig = plt.figure(figsize=size)
    else:
        fig = plt.figure()
    ax = fig.add_subplot(111)
    # axis limits and ticks
    ax.set_ylim(0, 1.05 * max(y))
    ax.set_xlim(-width / 2, len(ind))
    ax.set_xticks(ind + width / 2)
    xtick_names = ax.set_xticklabels(x)
    if l == 'gene':
        plt.setp(xtick_names, rotation=90, fontsize=7)
    if grid:
        ax.yaxis.grid(True, alpha=0.5)
    # axis labels
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    if xfontsize:
        plt.setp(xtick_names, fontsize=xfontsize)
    ax.tick_params(axis='x', which='both', top='off', length=3, pad=1.5)
    ax.tick_params(axis='y', which='both', right='off', length=3, pad=1.5)
    # make the plot
    bar = ax.bar(ind, y, width, color=colors)
    fig.tight_layout()
    plt.savefig(ofile)
    plt.close()


def make_heatmap(df, ofile):
    sns.set()
    # set up plot, determine plot size
    h, w = df.shape
    f, ax = plt.subplots(figsize=(w / 1.75, h / 3))
    sns.heatmap(df,
                square=True,
                cmap='Blues',
                cbar=True,
                cbar_kws={'orientation': 'horizontal',
                          'fraction': 0.02,
                          'pad': 0.02,
                          'shrink': 0.675})
    # adjust labels
    ax.xaxis.tick_top()
    plt.xticks(rotation=90)
    f.tight_layout()
    # make the plot
    plt.savefig(ofile)
    plt.close()





def print_collection_info(collection):
    print('')
    print('')
    print('-' * 25)
    print(collection)
    print('-' * 25)
    print('')




def run(**kwargs):
    args = Args(**kwargs)
    main(args)


def run_standalone(args):
    logfile = args.log if args.log else os.path.join(args.output, 'abstats.log')
    log.setup_logging(logfile)
    global logger
    logger = log.get_logger('abstats')
    main(args)


def main(args):
    db = mongodb.get_db(args.db, ip=args.ip, port=args.port, user=args.user, password=args.password)
    for collection in mongodb.get_collections(db):
        print_collection_info(collection)
        seqs = query(db, collection, args.chain)
        if len(seqs) == 0:
            continue
        germline_plot(seqs, 'V', collection, args.output, args.var_plot, args.species, args.chain)
        if args.chain == 'heavy':
            germline_plot(seqs, 'D', collection, args.output, args.div_plot, args.species, args.chain)
        germline_plot(seqs, 'J', collection, args.output, args.join_plot, args.species, args.chain)
        cdr3_plot(seqs, collection, args.cdr3_plot, args.chain, args.output)
        vj_heatmap(seqs, collection, args.heatmap, args.species, args.chain, args.output)




if __name__ == '__main__':
    args = parse_args()
    main()



#!/usr/bin/env python
# filename: pipeline.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import glob
import os

from . import log


def initialize(log_file, project_dir=None, debug=False):
    '''
    Initializes an AbTools pipeline.

    Initialization includes printing the AbTools splash, setting up logging,
    creating the project directory, and logging both the project directory
    and the log location.

    Args:

        log_file (str): Path to the log file. Required.

        project_dir (str): Path to the project directory. If not provided,
            the project directory won't be created and the location won't be logged.

        debug (bool): If ``True``, the logging level will be set to ``logging.DEBUG``.
            Default is ``FALSE``, which logs at ``logging.INFO``.

    Returns:

        logger
    '''
    print_splash()
    log.setup_logging(log_file, print_log_location=False, debug=debug)
    logger = log.get_logger('pipeline')
    if project_dir is not None:
        make_dir(os.path.normpath(project_dir))
        logger.info('PROJECT DIRECTORY: {}'.format(project_dir))
        logger.info('')
    logger.info('LOG LOCATION: {}'.format(log_file))
    print('')
    return logger


def make_dir(d):
    '''
    Makes a directory, if it doesn't already exist.

    Args:

        d (str): Path to a directory.
    '''
    if not os.path.exists(d):
        os.makedirs(d)


def list_files(d, extension=None):
    '''
    Lists files in a given directory.

    Args:

        d (str): Path to a directory.

        extension (str): If supplied, only files that contain the
            specificied extension will be returned. Default is ``False``,
            which returns all files in ``d``.

    Returns:

        list: A sorted list of file paths.
    '''
    if os.path.isdir(d):
        expanded_dir = os.path.expanduser(d)
        files = sorted(glob.glob(expanded_dir + '/*'))
    else:
        files = [d, ]
    if extension is not None:
        if type(extension) in [str, unicode]:
            extension = [extension, ]
        files = [f for f in files if any([f.split('.')[-1] in extension,
                                          f.split('.')[-1].upper() in extension,
                                          f.split('.')[-1].lower() in extension])]
    return files


def print_splash():
    splash = '''
     _    _   _____           _       ____  _            _ _            
    / \  | |_|_   _|__   ___ | |___  |  _ \(_)_ __   ___| (_)_ __   ___ 
   / _ \ | '_ \| |/ _ \ / _ \| / __| | |_) | | '_ \ / _ \ | | '_ \ / _ \\
  / ___ \| |_) | | (_) | (_) | \__ \ |  __/| | |_) |  __/ | | | | |  __/
 /_/   \_\_.__/|_|\___/ \___/|_|___/ |_|   |_| .__/ \___|_|_|_| |_|\___|
                                             |_|                        '''
    # print('')
    print(splash)
    print('')



#!/usr/bin/python
# filename: celery.py


###########################################################################
#
# Copyright (c) 2014 Bryan Briney.  All rights reserved.
#
# @version: 1.0.0
# @author: Bryan Briney
# @license: MIT (http://opensource.org/licenses/MIT)
#
###########################################################################



from __future__ import absolute_import

from celery import Celery


# instantiate Celery object
celery = Celery(include=['queue.tasks',
                         'abstar.utils.vdj',
                         'clonify.clonify.clonify_map',
                         'clonify.clonify.clonify_reduce', ])


# import celery config file
celery.config_from_object('abtools.queue.celeryconfig')


if __name__ == '__main__':
    celery.start()



#!/usr/bin/python
# filename: ab_phylogeny.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import subprocess as sp
import sys

import matplotlib as mpl
import seaborn as sns

from Bio import SeqIO

from abtools.ab_phylogeny.utils import msa, tree
from abtools.ab_phylogeny.utils.timepoint import Timepoint



def parse_args():
	import argparse
	parser = argparse.ArgumentParser("")
	parser.add_argument('-i', '--input', dest='input', required=True,
						help="Input file, containing sequences in FASTA format with the timepoint appended to the sequence ID. \
						Required.")
	parser.add_argument('-o', '--output', dest='output', required=True,
						help="Output directory for the FASTA, alignment, tree and figure files. \
						Required")
	parser.add_argument('-r', '--root', dest='root', required=True,
						help="FASTA file containing a single sequence that will be used to root the phylogenetic tree. \
						Required.")
	parser.add_argument('-m', '--mabs', dest='mabs', default=None,
						help="FASTA file containing sequences of monoclonal antibodies of the appropriate chain.")
	parser.add_argument('-a', '--alignment', dest='alignment', default=None,
						help="Provide a pre-aligned dataset. \
						Ensure that the root sequence is labeled as 'root' and that all mAb sequences are pre-pended with 'mab_'. \
						If provided, -i, -r and -m are ignored.")
	parser.add_argument('-n', '--newick', dest='newick', default=None,
						help="Provide a pre-generated Newick tree file. \
						Ensure that the root sequence is labeled as 'root' and that all mAb sequences are pre-pended with 'mab_'. \
						If provided -i, -r, -m and -a are ignored.")
	parser.add_argument('-s', '--sample_id', dest='sample_id', default=None,
						help="Sample ID. If not provided, the sample ID will be inferred from the input file.")
	parser.add_argument('-t', '--timepoints', dest='timepoints', default=None,
						help="Tab-delimited file, of the following format (one per line): \
						TimepointName	TimepointOrder	TimepointColor\
						TimepointName is the name appended to the sequences in the input file.\
						TimepointOrder is an integer that indicates the order in which the timepoints should be sorted.\
						TimepointColor is an RGB or hex value that will be used to color the phylogenetic tree.\
						If mAb sequences are provided, the 'mab' entry will be used to sort/color the mAb sequences.\
						If not provided, colors will be automatically selected and timepoints will be determined by a simple \
						sort of the raw timepoint values parsed from the input file.")
	parser.add_argument('-A', '--aa', dest='is_aa', default=False, action='store_true',
						help="If used, all input files contain amino acid sequences. \
						Default is nucleotide sequences.")
	parser.add_argument('-D', '--delimiter', dest='delimiter', default='_',
						help="Delimiter that separates the timepoint and sequence ID. \
						Cannot use ':', ';' or '=', since these can screw up the tree file. \
						Default is '_'.")
	parser.add_argument('-S', '--scale', dest='scale', default=None,
						help="Scale for the resulting ete2 tree. \
						If not provided, the ete2 default value will be used.")
	parser.add_argument('-b', '--branch_vertical_margin', dest='branch_vertical_margin', default=None,
						help="Branch vertical margin for the resulting ete2 tree. \
						If not provided, the ete2 default value will be used..")
	return parser.parse_args()


class Args(object):
	"""docstring for Args"""
	def __init__(self, input=None, output=None,
				 root=None, mabs=None,
				 alignment=None, newick=None,
				 sample_id=None, is_aa=False,
				 timepoints=None, timepoint_delimiter='_',
				 tree_scale=None, branch_vertical_margin=None):
		super(Args, self).__init__()
		if not all([input, output]):
			print('Input and Output files must be provided.')
			sys.exit(1)
		self.input = input
		self.output = output
		self.root = root
		self.mabs = mabs
		self.alignment = alignment
		self.newick = newick
		self.sample_id = sample_id
		self.timepoints = timepoints
		self.is_aa = is_aa
		self.timepoint_delimiter = timepoint_delimiter



# ================================================
#
#                 INPUT PARSING
#
# ================================================



def parse_seqs(args):
	seqs = parse_input_file(args.input)
	if args.root:
		seqs += parse_root(args.root)
	if args.mabs:
		seqs += parse_mabs(args.mabs, args.delimiter)
	timepoints = list(set([s.id.split(args.delimiter)[0] for s in seqs]))
	return seqs, timepoints


def parse_input_file(input_file):
	seqs = []
	for seq in SeqIO.parse(open(input_file, 'r'), 'fasta'):
		seqs.append(seq)
	return seqs


def parse_root(root_file):
	root = SeqIO.read(open(root_file, 'r'), 'fasta')
	root.id = 'root'
	return [root, ]


def parse_mabs(mabs_file, delimiter):
	seqs = []
	for seq in SeqIO.parse(open(mabs_file, 'r'), 'fasta'):
		if seq.id.split(delimiter)[0] != 'mab':
			seq.id = 'mab{}{}'.format(delimiter, seq.id)
		seqs.append(seq)
	return seqs


def parse_timepoints(tps, args):
	timepoints = []
	if args.timepoints:
		with open(args.timepoints, 'r') as f:
			for line in f:
				name, order, color = line.strip().split('\t')
				if name in tps:
					timepoints.append(Timepoint(name, order, color))
	else:
		colors = sns.hls_palette(len(tps), l=0.5, s=0.9)
		timepoints.append(Timepoint('root', 0, colors[0]))
		for i, tp in enumerate(sorted([t for t in tps if t != 'root'])):
			timepoints.append(Timepoint(tp, i + 1, colors[i + 1]))
	return timepoints


def make_msa(seqs, args):
	if args.sample_id:
		sample = args.sample_id
	else:
		sample = '.'.join(os.path.basename(args.input).split('.')[:-1])
	fasta_file = os.path.join(args.output, '{}.fasta'.format(sample))
	alignment = msa.align(seqs, fasta_file)
	return alignment


def make_tree(alignment, timepoints, args):
	return tree.make_tree(alignment,
						  timepoints,
						  args.delimiter,
						  args.is_aa,
						  args.scale,
						  args.branch_vertical_margin)


def run(**kwargs):
	args = Args(**kwargs)
	main(args)


def main(args):
	seqs, tps = parse_seqs(args)
	timepoints = parse_timepoints(tps, args)
	alignment = make_msa(seqs, args)
	make_tree(alignment, timepoints, args)


if __name__ == '__main__':
	args = parse_args()
	main(args)
